"""Command line argument parsing for cwltool."""

import argparse
import os
import re
import sys
import urllib
from collections.abc import Callable, MutableSequence, Sequence
from typing import Any, cast

import rich.markup
from rich.table import Table
from rich_argparse import HelpPreviewAction, RichHelpFormatter
from typing_extensions import LiteralString

from .loghandler import _logger
from .process import Process, shortname
from .resolver import ga4gh_tool_registries
from .software_requirements import SOFTWARE_REQUIREMENTS_ENABLED
from .utils import DEFAULT_TMP_PREFIX


class _env_var_table(Table):
    def __init__(self) -> None:
        super().__init__(show_edge=False)
        self.defs: list[tuple[str, str]] = []
        self.add_column("environment variable")
        self.add_column("usage")

    def add_env(self, name: str, description: str) -> None:
        self.defs.append((name, description))
        self.add_row(name, description)

    def splitlines(self) -> list[LiteralString]:  # used by sphinx.autoprogram
        markup_link = re.compile(r"\[link=(.*?)](.*?)\[/]")
        owh = RichHelpFormatter.styles["argparse.args"]  # options with hyphens
        bt = RichHelpFormatter.styles["argparse.syntax"]  # backticks
        owh_regex = re.compile(rf"\[{owh}](.*?)\[/]")
        bt_regex = re.compile(rf"\[{bt}](.*?)\[/]")
        lines = ["environment variables"]
        lines.append("!" * len(lines[0]))

        def _cvt(line: str) -> str:
            result = re.sub(markup_link, r"`\2 <\1>`__", line)
            result = re.sub(owh_regex, r":option:`\1`", result)
            result = re.sub(bt_regex, r"``\1``", result)
            return result

        for name, description in self.defs:
            lines.append(f".. envvar:: {name}")
            lines.append("")
            lines.append(f"    {_cvt(description)}")
            lines.append("")
        return lines


def arg_parser() -> argparse.ArgumentParser:
    RichHelpFormatter.group_name_formatter = str
    owh = RichHelpFormatter.styles["argparse.args"]  # options with hyphens
    bt = RichHelpFormatter.styles["argparse.syntax"]  # backticks
    env_table = _env_var_table()
    env_table.add_env("CWLTOOL_OPTIONS", f"set additional space-separated [{bt}]cwltool[/] options")
    env_table.add_env(
        "CWL_FULL_NAME", f"full name of user to record as part of [{owh}]--provenance[/]"
    )
    env_table.add_env(
        "CWL_SINGULARITY_CACHE",
        f"directory path to find already downloaded Singularity images; "
        f"[{bt}]dockerFile[/] images will also be searched for and stored here",
    )
    env_table.add_env(
        "ORCID",
        f"user [link=https://orcid.org/]ORCID identifier[/] to record as part of [{owh}]--provenance[/]",
    )
    env_table.add_env(
        "XDG_CACHE_HOME",
        f"directory path to store cached HTTP(S) downloads in a [{bt}]cwltool[/] sub-folder; "
        f"if unset, then cache downloads are stored in [{bt}]$HOME/.cache/cwltool[/]",
    )
    env_table.add_env(
        "XDG_DATA_HOME",
        f"directory path to search for a [{bt}].local/commonwl[/] "
        f"CWL tool folder; if unset, then [{bt}]$HOME/.local/commonwl[/] is searched",
    )
    env_table.add_env(
        "XDG_DATA_DIRS",
        f"directory path to search for a [{bt}]commonwl[/] "
        f"CWL tool folder; if unset, then [{bt}]/usr/local/share/[/] and [{bt}]/usr/share/[/] are searched",
    )

    parser = argparse.ArgumentParser(
        formatter_class=RichHelpFormatter,
        description="Reference executor for Common Workflow Language standards. "
        "Not for production use.",
        epilog=env_table,  # type: ignore[arg-type]
    )
    if sys.version_info >= (3, 14):
        parser.suggest_on_error = True

    parser.add_argument("--basedir", type=str)
    parser.add_argument(
        "--outdir",
        type=str,
        default=os.path.abspath("."),
        help="Output directory. The default is the current directory.",
    )
    parser.add_argument(
        "--write-summary",
        "-w",
        type=str,
        help="Path to write the final output JSON object to. Default is stdout.",
        default="",
        dest="write_summary",
    )
    parser.add_argument(
        "--parallel",
        action="store_true",
        default=False,
        help="Run jobs in parallel. ",
    )
    parser.add_argument(
        "--parallel-max",
        type=int,
        default=0,
        help="Maximum number of jobs to run in parallel. "
        "Specify '0' to match the number of CPU cores available.",
    )
    parser.add_argument(
        "--skip-schemas",
        action="store_true",
        help="Skip loading of schemas",
        default=False,
        dest="skip_schemas",
    )
    parser.add_argument(
        "--enable-dev",
        action="store_true",
        help="Enable loading and running unofficial development versions of " "the CWL standards.",
        default=False,
    )
    parser.add_argument(
        "--enable-ext",
        action="store_true",
        help="Enable loading and running `cwltool:` extensions to the CWL standards.",
        default=False,
    )
    parser.add_argument(
        "--disable-validate",
        dest="do_validate",
        action="store_false",
        default=True,
        help=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--fast-parser",
        dest="fast_parser",
        action="store_true",
        default=False,
        help=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--on-error",
        help="Desired workflow behavior when a step fails.  One of `stop` (do "
        "not submit any more steps) or `continue` (may submit other steps that "
        "are not downstream from the error). Default is `stop`.",
        default="stop",
        choices=("stop", "continue"),
    )
    parser.add_argument(
        "--relax-path-checks",
        action="store_true",
        default=False,
        help="Relax requirements on path names to permit " "spaces and hash characters.",
        dest="relax_path_checks",
    )
    parser.add_argument(
        "--overrides",
        type=str,
        default=None,
        help="Read process requirement overrides from file.",
    )
    parser.add_argument(
        "--mpi-config-file",
        type=str,
        default=None,
        help="Platform specific configuration for MPI (parallel launcher, its "
        "flag etc). See README section 'Running MPI-based tools' for details "
        "of the format.",
    )
    parser.add_argument(
        "--generate-help-preview",
        action=HelpPreviewAction,
        path="help-preview.svg",  # (optional) or "help-preview.html" or "help-preview.txt"
    )
    parser.add_argument(
        "workflow",
        type=str,
        nargs="?",
        default=None,
        metavar="cwl_document",
        help="path or URL to a CWL Workflow, "
        "CommandLineTool, or ExpressionTool. If the `inputs_object` has a "
        "`cwl:tool` field indicating the path or URL to the cwl_document, "
        " then the `cwl_document` argument is optional.",
    )
    parser.add_argument(
        "job_order",
        nargs=argparse.REMAINDER,
        metavar="inputs_object",
        help="path or URL to a YAML or JSON "
        "formatted description of the required input values for the given "
        "`cwl_document`.",
    )

    non_exec_group = parser.add_argument_group(title="non-execution options")
    printgroup = non_exec_group.add_mutually_exclusive_group()
    printgroup.add_argument(
        "--print-rdf",
        action="store_true",
        help="Print corresponding RDF graph for workflow and exit",
    )
    printgroup.add_argument(
        "--print-dot",
        action="store_true",
        help="Print workflow visualization in graphviz format and exit",
    )
    printgroup.add_argument(
        "--print-pre",
        action="store_true",
        help="Print CWL document after preprocessing.",
    )
    printgroup.add_argument(
        "--print-deps", action="store_true", help="Print CWL document dependencies."
    )
    printgroup.add_argument(
        "--print-input-deps",
        action="store_true",
        help="Print input object document dependencies.",
    )
    printgroup.add_argument(
        "--pack",
        action="store_true",
        help="Combine components into single document and print.",
    )
    printgroup.add_argument("--version", action="store_true", help="Print version and exit")
    printgroup.add_argument("--validate", action="store_true", help="Validate CWL document only.")
    printgroup.add_argument(
        "--print-supported-versions",
        action="store_true",
        help="Print supported CWL specs.",
    )
    printgroup.add_argument(
        "--print-subgraph",
        action="store_true",
        help="Print workflow subgraph that will execute. Can combined with "
        "--target or --single-step",
    )
    printgroup.add_argument(
        "--print-targets", action="store_true", help="Print targets (output parameters)"
    )
    printgroup.add_argument(
        "--make-template", action="store_true", help="Generate a template input object"
    )
    non_exec_group.add_argument(
        "--relative-deps",
        choices=["primary", "cwd"],
        default="primary",
        help="When using --print-deps, print paths "
        "relative to primary file or current working directory.",
    )
    non_exec_group.add_argument(
        "--rdf-serializer",
        help="Output RDF serialization format used by --print-rdf (one of "
        "turtle (default), n3, nt, xml)",
        default="turtle",
    )
    non_exec_group.add_argument(
        "--tool-help", action="store_true", help="Print command line help for tool"
    )

    logging_group = parser.add_argument_group(title="logging configuration")
    logging_exclusive = logging_group.add_mutually_exclusive_group()
    logging_exclusive.add_argument("--verbose", action="store_true", help="Default logging")
    logging_exclusive.add_argument("--no-warnings", action="store_true", help="Only print errors.")
    logging_exclusive.add_argument(
        "--quiet", action="store_true", help="Only print warnings and errors."
    )
    logging_exclusive.add_argument("--debug", action="store_true", help="Print even more logging")
    logging_group.add_argument(
        "--log-dir",
        type=str,
        default="",
        help="Log your tools stdout/stderr to this location outside of container "
        "This will only log stdout/stderr if you specify stdout/stderr in their "
        "respective fields or capture it as an output",
    )
    logging_group.add_argument(
        "--timestamps",
        action="store_true",
        help="Add timestamps to the errors, warnings, and notifications.",
    )

    envgroup = parser.add_argument_group(title="environment variables control")
    env_exclusive = envgroup.add_mutually_exclusive_group()
    env_exclusive.add_argument(
        "--preserve-environment",
        type=str,
        action="append",
        help="Preserve specific environment variable when running "
        "CommandLineTools. May be provided multiple times. By default PATH is "
        "preserved when not running in a container.",
        metavar="ENVVAR",
        default=[],
        dest="preserve_environment",
    )
    env_exclusive.add_argument(
        "--preserve-entire-environment",
        action="store_true",
        help="Preserve all environment variables when running CommandLineTools "
        "without a software container.",
        default=False,
        dest="preserve_entire_environment",
    )

    files_group = parser.add_argument_group(
        title="intermediate, temporary, or final output files management"
    )
    files_group.add_argument(
        "--tmpdir-prefix",
        type=str,
        help="Path prefix for temporary directories. If --tmpdir-prefix is not "
        "provided, then the prefix for temporary directories is influenced by "
        "the value of the TMPDIR, TEMP, or TMP environment variables. Taking "
        "those into consideration, the current default is {}.".format(DEFAULT_TMP_PREFIX),
        default=DEFAULT_TMP_PREFIX,
    )

    intgroup = files_group.add_mutually_exclusive_group()
    intgroup.add_argument(
        "--tmp-outdir-prefix",
        type=str,
        help="Path prefix for intermediate output directories. Defaults to the "
        "value of --tmpdir-prefix.",
        default="",
    )
    intgroup.add_argument(
        "--cachedir",
        type=str,
        default="",
        help="Directory to cache intermediate workflow outputs to avoid "
        "recomputing steps. Can be very helpful in the development and "
        "troubleshooting of CWL documents.",
    )

    tmpgroup = files_group.add_mutually_exclusive_group()
    tmpgroup.add_argument(
        "--rm-tmpdir",
        action="store_true",
        default=True,
        help="Delete intermediate temporary directories (default)",
        dest="rm_tmpdir",
    )
    tmpgroup.add_argument(
        "--leave-tmpdir",
        action="store_false",
        default=True,
        help="Do not delete intermediate temporary directories",
        dest="rm_tmpdir",
    )

    outgroup = files_group.add_mutually_exclusive_group()
    outgroup.add_argument(
        "--move-outputs",
        action="store_const",
        const="move",
        default="move",
        help="Move output files to the workflow output directory and delete "
        "intermediate output directories (default).",
        dest="move_outputs",
    )
    outgroup.add_argument(
        "--leave-outputs",
        action="store_const",
        const="leave",
        default="move",
        help="Leave output files in intermediate output directories.",
        dest="move_outputs",
    )
    outgroup.add_argument(
        "--copy-outputs",
        action="store_const",
        const="copy",
        default="move",
        help="Copy output files to the workflow output directory and don't "
        "delete intermediate output directories.",
        dest="move_outputs",
    )

    provgroup = parser.add_argument_group("provenance recording")
    provgroup.add_argument(
        "--provenance",
        help="Save provenance to specified folder as a "
        "Research Object that captures and aggregates "
        "workflow execution and data products.",
        type=str,
    )
    provgroup.add_argument(
        "--enable-user-provenance",
        default=False,
        action="store_true",
        help="Record user account info as part of --provenance.",
        dest="user_provenance",
    )
    provgroup.add_argument(
        "--disable-user-provenance",
        default=False,
        action="store_false",
        help="Do not record user account info in --provenance.",
        dest="user_provenance",
    )
    provgroup.add_argument(
        "--enable-host-provenance",
        default=False,
        action="store_true",
        help="Record host info as part of --provenance.",
        dest="host_provenance",
    )
    provgroup.add_argument(
        "--disable-host-provenance",
        default=False,
        action="store_false",
        help="Do not record host info in --provenance.",
        dest="host_provenance",
    )
    provgroup.add_argument(
        "--orcid",
        help="Record user ORCID identifier as part of "
        "--provenance, e.g. https://orcid.org/0000-0002-1825-0097 "
        "or 0000-0002-1825-0097. Alternatively the environment variable "
        "ORCID may be set.",
        dest="orcid",
        default=os.environ.get("ORCID", ""),
        type=str,
    )
    provgroup.add_argument(
        "--full-name",
        help="Record full name of user as part of --provenance, "
        "e.g. `Josiah Carberry`. You may need to use shell quotes to preserve "
        "spaces. Also set by the CWL_FULL_NAME environment variable.",
        dest="cwl_full_name",
        default=os.environ.get("CWL_FULL_NAME", ""),
        type=str,
    )

    strictgroup = parser.add_mutually_exclusive_group()
    strictgroup.add_argument(
        "--strict",
        action="store_true",
        help="Strict validation (unrecognized or out of place fields are errors)",
        default=True,
        dest="strict",
    )
    strictgroup.add_argument(
        "--non-strict",
        action="store_false",
        help="Lenient validation (ignore unrecognized fields)",
        default=True,
        dest="strict",
    )

    doccachegroup = parser.add_mutually_exclusive_group()
    doccachegroup.add_argument(
        "--no-doc-cache",
        action="store_false",
        help="Disable disk cache for documents loaded over HTTP",
        default=True,
        dest="doc_cache",
    )
    doccachegroup.add_argument(
        "--doc-cache",
        action="store_true",
        help="Enable disk cache for documents loaded over HTTP",
        default=True,
        dest="doc_cache",
    )

    container_group = parser.add_argument_group(
        title="software container engine selection and configuration"
    )
    dockergroup = container_group.add_mutually_exclusive_group()
    dockergroup.add_argument(
        "--user-space-docker-cmd",
        metavar="CMD",
        help="Specify the path to udocker. Implies --udocker",
    )
    dockergroup.add_argument(
        "--udocker",
        help="Use the udocker runtime for running containers "
        "(equivalent to --user-space-docker-cmd=udocker).",
        action="store_const",
        const="udocker",
        dest="user_space_docker_cmd",
    )
    dockergroup.add_argument(
        "--singularity",
        action="store_true",
        default=False,
        help="Use "
        "Singularity or Apptainer runtime for running containers. "
        "Requires Singularity v2.6.1+ and Linux with kernel "
        "version v3.18+ or with overlayfs support "
        "backported.",
    )
    dockergroup.add_argument(
        "--podman",
        action="store_true",
        default=False,
        help="Use Podman runtime for running containers. ",
    )
    dockergroup.add_argument(
        "--no-container",
        action="store_false",
        default=True,
        help="Do not execute jobs in a "
        "Docker container, even when `DockerRequirement` "
        "is specified under `hints`.",
        dest="use_container",
    )

    pullgroup = container_group.add_mutually_exclusive_group()
    pullgroup.add_argument(
        "--enable-pull",
        default=True,
        action="store_true",
        help="Try to pull Docker images",
        dest="pull_image",
    )
    pullgroup.add_argument(
        "--disable-pull",
        default=True,
        action="store_false",
        help="Do not try to pull Docker images",
        dest="pull_image",
    )

    container_group.add_argument(
        "--singularity-sandbox-path",
        default=None,
        type=str,
        help="Singularity/Apptainer sandbox image base path. "
        "Will use a pre-existing sandbox image. "
        "Will be prepended to the dockerPull path. "
        "Equivalent to use CWL_SINGULARITY_IMAGES variable. ",
        dest="image_base_path",
    )
    container_group.add_argument(
        "--force-docker-pull",
        action="store_true",
        default=False,
        help="Pull latest software container image even if it is locally present",
        dest="force_docker_pull",
    )
    container_group.add_argument(
        "--no-read-only",
        action="store_true",
        default=False,
        help="Do not set root directory in the container as read-only",
        dest="no_read_only",
    )
    container_group.add_argument(
        "--default-container",
        help="Specify a default software container to use for any "
        "CommandLineTool without a DockerRequirement.",
    )
    container_group.add_argument(
        "--no-match-user",
        action="store_true",
        help="Disable passing the current uid to `docker run --user`",
    )
    container_group.add_argument(
        "--custom-net",
        type=str,
        help="Passed to `docker run` as the `--net` parameter when "
        "NetworkAccess is true, which is its default setting.",
    )
    container_group.add_argument(
        "--strict-memory-limit",
        action="store_true",
        help="When running with "
        "software containers and the Docker engine, pass either the "
        "calculated memory allocation from ResourceRequirements or the "
        "default of 1 gigabyte to Docker's --memory option.",
    )
    container_group.add_argument(
        "--strict-cpu-limit",
        action="store_true",
        help="When running with "
        "software containers and the Docker engine, pass either the "
        "calculated cpu allocation from ResourceRequirements or the "
        "default of 1 core to Docker's --cpu option. "
        "Requires docker version >= v1.13.",
    )

    container_cleanup = container_group.add_mutually_exclusive_group()
    container_cleanup.add_argument(
        "--rm-container",
        action="store_true",
        default=True,
        help="Delete Docker container used by jobs after they exit (default)",
        dest="rm_container",
    )
    container_cleanup.add_argument(
        "--leave-container",
        action="store_false",
        default=True,
        help="Do not delete Docker container used by jobs after they exit",
        dest="rm_container",
    )
    container_group.add_argument(
        # Disabled as containerid is now saved by default
        "--record-container-id",
        action="store_true",
        default=False,
        help=argparse.SUPPRESS,
        dest="record_container_id",
    )
    container_group.add_argument(
        "--cidfile-dir",
        type=str,
        help="Store the software container ID into a file in the specified directory.",
        default=None,
        dest="cidfile_dir",
    )
    container_group.add_argument(
        "--cidfile-prefix",
        type=str,
        help="Specify a prefix to the software container ID filename. "
        "Final file name will be followed by a timestamp. "
        "The default is no prefix.",
        default=None,
        dest="cidfile_prefix",
    )

    dep_resolver_group = parser.add_argument_group("dependency resolver configuration")
    dependency_resolvers_configuration_help = argparse.SUPPRESS
    dependencies_directory_help = argparse.SUPPRESS
    use_biocontainers_help = argparse.SUPPRESS
    conda_dependencies = argparse.SUPPRESS

    if SOFTWARE_REQUIREMENTS_ENABLED:
        dependency_resolvers_configuration_help = (
            "Dependency resolver "
            "configuration file describing how to adapt `SoftwareRequirement` "
            "packages to current system."
        )
        dependencies_directory_help = (
            "Default root directory used by dependency resolvers configuration."
        )
        use_biocontainers_help = (
            "Use biocontainers for tools without an " "explicitly annotated Docker container."
        )
        conda_dependencies = "Short cut to use Conda to resolve `SoftwareRequirement` packages."

    dep_resolver_group.add_argument(
        "--beta-dependency-resolvers-configuration",
        default=None,
        help=dependency_resolvers_configuration_help,
    )
    dep_resolver_group.add_argument(
        "--beta-dependencies-directory", default=None, help=dependencies_directory_help
    )
    dep_resolver_group.add_argument(
        "--beta-use-biocontainers",
        default=None,
        help=use_biocontainers_help,
        action="store_true",
    )
    dep_resolver_group.add_argument(
        "--beta-conda-dependencies",
        default=None,
        help=conda_dependencies,
        action="store_true",
    )

    colorgroup = parser.add_mutually_exclusive_group()
    colorgroup.add_argument(
        "--enable-color",
        action="store_true",
        help="Enable logging color (default enabled)",
        default=True,
    )
    colorgroup.add_argument(
        "--disable-color",
        action="store_false",
        dest="enable_color",
        help="Disable colored logging (default false)",
    )

    reg_api_group = parser.add_argument_group(
        title="tool resolution", description="(using the GA4GH Tool Registry Service (TRS) API)"
    )
    reg_exclusive = reg_api_group.add_mutually_exclusive_group()
    reg_exclusive.add_argument(
        "--enable-ga4gh-tool-registry",
        action="store_true",
        help="Enable tool resolution using GA4GH TRS API",
        dest="enable_ga4gh_tool_registry",
        default=True,
    )
    reg_exclusive.add_argument(
        "--disable-ga4gh-tool-registry",
        action="store_false",
        help="Disable tool resolution using GA4GH TRS API",
        dest="enable_ga4gh_tool_registry",
        default=True,
    )
    reg_api_group.add_argument(
        "--add-ga4gh-tool-registry",
        action="append",
        help="Add a GA4GH TRS endpoint to use for resolution, default %s" % ga4gh_tool_registries,
        dest="ga4gh_tool_registries",
        default=[],
    )

    checkgroup = parser.add_mutually_exclusive_group()
    checkgroup.add_argument(
        "--compute-checksum",
        action="store_true",
        default=True,
        help="Compute checksum of contents while collecting outputs",
        dest="compute_checksum",
    )
    checkgroup.add_argument(
        "--no-compute-checksum",
        action="store_false",
        help="Do not compute checksum of contents while collecting outputs",
        dest="compute_checksum",
    )

    jsgroup = parser.add_argument_group(title="javascript options")
    jsgroup.add_argument(
        "--eval-timeout",
        help="Time to wait for a Javascript expression to evaluate before giving "
        "an error, default 60s.",
        type=float,
        default=60,
    )
    jsgroup.add_argument(
        "--js-console", action="store_true", help="Enable javascript console output"
    )
    jsgroup.add_argument(
        "--disable-js-validation",
        action="store_true",
        help="Disable javascript validation.",
    )
    jsgroup.add_argument(
        "--js-hint-options-file",
        type=str,
        help="File of options to pass to jshint. "
        'This includes the added option "includewarnings". ',
    )

    target_group = parser.add_argument_group(
        title="execution target selection",
        description="(optional, by default all target contributing to final outputs are executed)",
    )
    subgroup = target_group.add_mutually_exclusive_group()
    subgroup.add_argument(
        "--target",
        "-t",
        action="append",
        help="Only execute steps that contribute to listed targets (can be "
        "provided more than once).",
    )
    subgroup.add_argument(
        "--single-step",
        type=str,
        default=None,
        help="Only executes a single step in a workflow. The input object must "
        "match that step's inputs. Can be combined with --print-subgraph.",
    )
    subgroup.add_argument(
        "--single-process",
        type=str,
        default=None,
        help="Only executes the underlying Process (CommandLineTool, "
        "ExpressionTool, or sub-Workflow) for the given step in a workflow. "
        "This will not include any step-level processing: `scatter`, `when`; "
        "and there will be no processing of step-level `default`, or `valueFrom` "
        "input modifiers. However, requirements/hints from the step or parent "
        "workflow(s) will be inherited as usual."
        "The input object must match that Process's inputs.",
    )

    return parser


def get_default_args() -> dict[str, Any]:
    """Get default values of cwltool's command line options."""
    ap = arg_parser()
    args = ap.parse_args([])
    return vars(args)


class FSAction(argparse.Action):
    """Base action for our custom actions."""

    objclass: str | None = None

    def __init__(
        self,
        option_strings: list[str],
        dest: str,
        nargs: Any = None,
        urljoin: Callable[[str, str], str] = urllib.parse.urljoin,
        base_uri: str = "",
        **kwargs: Any,
    ) -> None:
        """Fail if nargs is used."""
        if nargs is not None:
            raise ValueError("nargs not allowed")
        self.urljoin = urljoin
        self.base_uri = base_uri
        super().__init__(option_strings, dest, **kwargs)

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        setattr(
            namespace,
            self.dest,
            {
                "class": self.objclass,
                "location": self.urljoin(self.base_uri, cast(str, values)),
            },
        )


class FSAppendAction(argparse.Action):
    """Appending version of the base action for our custom actions."""

    objclass: str | None = None

    def __init__(
        self,
        option_strings: list[str],
        dest: str,
        nargs: Any = None,
        urljoin: Callable[[str, str], str] = urllib.parse.urljoin,
        base_uri: str = "",
        **kwargs: Any,
    ) -> None:
        """Initialize."""
        if nargs is not None:
            raise ValueError("nargs not allowed")
        self.urljoin = urljoin
        self.base_uri = base_uri
        super().__init__(option_strings, dest, **kwargs)

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        g = getattr(namespace, self.dest)
        if not g:
            g = []
            setattr(namespace, self.dest, g)
        g.append(
            {
                "class": self.objclass,
                "location": self.urljoin(self.base_uri, cast(str, values)),
            }
        )


class FileAction(FSAction):
    objclass: str | None = "File"


class DirectoryAction(FSAction):
    objclass: str | None = "Directory"


class FileAppendAction(FSAppendAction):
    objclass: str | None = "File"


class DirectoryAppendAction(FSAppendAction):
    objclass: str | None = "Directory"


class AppendAction(argparse.Action):
    """An argparse action that clears the default values if any value is provided."""

    _called: bool
    """Initially set to ``False``, changed if any value is appended."""

    def __init__(
        self,
        option_strings: list[str],
        dest: str,
        nargs: Any = None,
        **kwargs: Any,
    ) -> None:
        """Initialize."""
        super().__init__(option_strings, dest, **kwargs)
        self._called = False

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        g = getattr(namespace, self.dest, None)
        if g is None:
            g = []
        if self.default is not None and not self._called:
            # If any value was specified, we then clear the list of options before appending.
            # We cannot always clear the ``default`` attribute since it collects the ``values`` appended.
            self.default.clear()
            self._called = True
        g.append(values)
        setattr(namespace, self.dest, g)


def add_argument(
    toolparser: argparse.ArgumentParser,
    name: str,
    inptype: Any,
    records: list[str],
    description: str = "",
    default: Any = None,
    input_required: bool = True,
    urljoin: Callable[[str, str], str] = urllib.parse.urljoin,
    base_uri: str = "",
) -> None:
    description = rich.markup.escape(description)
    if len(name) == 1:
        flag = "-"
    else:
        flag = "--"

    # if input_required is false, don't make the command line
    # parameter required.
    required = default is None and input_required
    if isinstance(inptype, MutableSequence):
        if len(inptype) == 1:
            inptype = inptype[0]
        elif len(inptype) == 2 and inptype[0] == "null":
            required = False
            inptype = inptype[1]
        elif len(inptype) == 2 and inptype[1] == "null":
            required = False
            inptype = inptype[0]
        else:
            _logger.debug("Can't make command line argument from %s", inptype)
            return None

    ahelp = description.replace("%", "%%")
    action: type[argparse.Action] | str | None = None
    atype: Any | None = None
    typekw: dict[str, Any] = {}

    match inptype:
        case "File":
            action = FileAction
        case "Directory":
            action = DirectoryAction
        case {"type": "array", "items": "File"}:
            action = FileAppendAction
        case {"type": "array", "items": "Directory"}:
            action = DirectoryAppendAction
        case {"type": "array", "items": str(items)}:
            action = AppendAction
            match items:
                case "int" | "long":
                    atype = int
                case "double" | "float":
                    atype = float
        case {"type": "enum"}:
            atype = str
        case {"type": "record", "fields": list(fields)}:
            records.append(name)
            for field in fields:
                fieldname = name + "." + shortname(field["name"])
                fieldtype = field["type"]
                fielddescription = field.get("doc", "")
                add_argument(
                    toolparser,
                    fieldname,
                    fieldtype,
                    records,
                    fielddescription,
                    default=default.get(shortname(field["name"]), None) if default else None,
                    input_required=required,
                )
            return
        case "string":
            atype = str
        case "int" | "long":
            atype = int
        case "double" | "float":
            atype = float
        case "boolean":
            action = "store_true"
        case _:
            _logger.debug("Can't make command line argument from %s", inptype)
            return None

    if action in (FileAction, DirectoryAction, FileAppendAction, DirectoryAppendAction):
        typekw["urljoin"] = urljoin
        typekw["base_uri"] = base_uri

    if inptype != "boolean":
        typekw["type"] = atype

    toolparser.add_argument(
        flag + name,
        required=required,
        help=ahelp,
        action=action,  # type: ignore
        default=default,
        **typekw,
    )


def generate_parser(
    toolparser: argparse.ArgumentParser,
    tool: Process,
    namemap: dict[str, str],
    records: list[str],
    input_required: bool = True,
    urljoin: Callable[[str, str], str] = urllib.parse.urljoin,
    base_uri: str = "",
) -> argparse.ArgumentParser:
    """Generate an ArgumentParser for the given CWL Process."""
    toolparser.description = tool.tool.get("doc", tool.tool.get("label", None))
    toolparser.add_argument("job_order", nargs="?", help="Job input json file")
    namemap["job_order"] = "job_order"

    for inp in tool.tool["inputs"]:
        name = shortname(inp["id"])
        namemap[name.replace("-", "_")] = name
        inptype = inp["type"]
        description = inp.get("doc", inp.get("label", ""))
        default = inp.get("default", None)
        add_argument(
            toolparser,
            name,
            inptype,
            records,
            description,
            default,
            input_required,
            urljoin,
            base_uri,
        )

    toolparser.add_argument(
        "--generate-help-preview",
        action=HelpPreviewAction,
        path="help-preview.svg",  # (optional) or "help-preview.html" or "help-preview.txt"
    )

    return toolparser



from collections.abc import Iterator
from typing import Optional

from .base import Client

def build(
    self: Client,
    recipe: str | None = ...,
    image: str | None = ...,
    isolated: bool | None = ...,
    sandbox: bool | None = ...,
    writable: bool | None = ...,
    build_folder: str | None = ...,
    robot_name: bool | None = ...,
    ext: str | None = ...,
    sudo: bool | None = ...,
    stream: bool | None = ...,
    force: bool | None = ...,
    options: list[str] | None | None = ...,
    quiet: bool | None = ...,
    return_result: bool | None = ...,
    sudo_options: str | list[str] | None = ...,
    singularity_options: list[str] | None = ...,
) -> tuple[str, Iterator[str]]: ...



import json

from .util import get_data, get_main_output


def test_cwl_output_json_missing_field_v1_0() -> None:
    """Confirm that unknown outputs are propagated from cwl.output.json in CWL v1.0."""
    err_code, stdout, _ = get_main_output([get_data("tests/test-cwl-out-v1.0.cwl")])
    assert err_code == 0
    assert "foo" in json.loads(stdout)


def test_cwl_output_json_missing_field_v1_1() -> None:
    """Confirm that unknown outputs are propagated from cwl.output.json in CWL v1.1."""
    err_code, stdout, _ = get_main_output([get_data("tests/test-cwl-out-v1.1.cwl")])
    assert err_code == 0
    assert "foo" in json.loads(stdout)


def test_cwl_output_json_missing_field_v1_2() -> None:
    """Confirm that unknown outputs are propagated from cwl.output.json in CWL v1.2."""
    err_code, stdout, _ = get_main_output([get_data("tests/test-cwl-out-v1.2.cwl")])
    assert err_code == 0
    assert "foo" in json.loads(stdout)


def test_cwl_output_json_missing_field_v1_3() -> None:
    """Confirm that unknown outputs are propagated from cwl.output.json in CWL v1.3."""
    err_code, stdout, stderr = get_main_output(
        ["--enable-dev", get_data("tests/test-cwl-out-v1.3.cwl")]
    )
    assert err_code == 0
    assert "foo" not in json.loads(stdout)
    assert "Discarded undeclared output named 'foo' from" in stderr



import copy
import itertools
import json
import logging
from collections.abc import MutableMapping, MutableSequence
from importlib.resources import files
from typing import Any, NamedTuple, cast

from cwl_utils.errors import SubstitutionError
from cwl_utils.expression import scanner as scan_expression
from cwl_utils.sandboxjs import code_fragment_to_js, exec_js_process
from ruamel.yaml.comments import CommentedMap, CommentedSeq
from schema_salad.avro.schema import (
    ArraySchema,
    EnumSchema,
    RecordSchema,
    Schema,
    UnionSchema,
)
from schema_salad.sourceline import SourceLine
from schema_salad.utils import json_dumps
from schema_salad.validate import validate_ex

from .errors import WorkflowException
from .loghandler import _logger


def is_expression(tool: Any, schema: Schema | None) -> bool:
    """Test a field/schema combo to see if it is a CWL Expression."""
    return (
        isinstance(schema, EnumSchema)
        and schema.name == "org.w3id.cwl.cwl.Expression"
        and isinstance(tool, str)
    )


class SuppressLog(logging.Filter):
    def __init__(self, name: str) -> None:
        """Initialize this log suppressor."""
        name = str(name)
        super().__init__(name)

    def filter(self, record: logging.LogRecord) -> bool:
        """Never accept a record."""
        return False


_logger_validation_warnings = logging.getLogger("cwltool.validation_warnings")
_logger_validation_warnings.addFilter(SuppressLog("cwltool.validation_warnings"))


def get_expressions(
    tool: CommentedMap | str | CommentedSeq,
    schema: Schema | ArraySchema | None,
    source_line: SourceLine | None = None,
) -> list[tuple[str, SourceLine | None]]:
    debug = _logger.isEnabledFor(logging.DEBUG)
    if is_expression(tool, schema):
        return [(cast(str, tool), source_line)]
    match schema:
        case UnionSchema(schemas=schemas):
            valid_schema = None

            for possible_schema in schemas:
                if is_expression(tool, possible_schema):
                    return [(cast(str, tool), source_line)]
                elif validate_ex(
                    possible_schema,
                    tool,
                    raise_ex=False,
                    logger=_logger_validation_warnings,
                    vocab={},
                ):
                    valid_schema = possible_schema

            return get_expressions(tool, valid_schema, source_line)
        case ArraySchema(items=items):
            if not isinstance(tool, MutableSequence):
                return []

            return list(
                itertools.chain(
                    *map(
                        lambda x: get_expressions(
                            x[1], items, SourceLine(tool, x[0])  # noqa: B009
                        ),
                        enumerate(tool),
                    )
                )
            )
        case RecordSchema(fields=fields):
            if not isinstance(tool, MutableMapping):
                return []

            expression_nodes = []

            for schema_field in fields:
                if schema_field.name in tool:
                    expression_nodes.extend(
                        get_expressions(
                            tool[schema_field.name],
                            schema_field.type,
                            SourceLine(tool, schema_field.name, include_traceback=debug),
                        )
                    )
            return expression_nodes
        case _:
            return []


class JSHintJSReturn(NamedTuple):
    """List of errors and the final values of the globals from running javascript."""

    errors: list[str]
    globals: list[str]


def jshint_js(
    js_text: str,
    globals: list[str] | None = None,
    options: dict[str, list[str] | str | int] | None = None,
    container_engine: str = "docker",
    eval_timeout: float = 60,
) -> JSHintJSReturn:
    if globals is None:
        globals = []
    if options is None:
        options = {
            "includewarnings": [
                "W117",  # <VARIABLE> not defined
                "W104",
                "W119",  # using ES6 features
            ],
            "strict": "implied",
            "esversion": 5,
        }

    res = files("cwltool").joinpath("jshint/jshint.js")
    # NOTE: we need a global variable for lodash (which jshint depends on)
    jshint_functions_text = "var global = this;" + res.read_text("utf-8")

    res2 = files("cwltool").joinpath("jshint/jshint_wrapper.js")
    # NOTE: we need to assign to ob, as the expression {validateJS: validateJS} as an expression
    # is interpreted as a block with a label `validateJS`
    jshint_functions_text += (
        "\n" + res2.read_text("utf-8") + "\nvar ob = {validateJS: validateJS}; ob"
    )

    returncode, stdout, stderr = exec_js_process(
        "validateJS(%s)" % json_dumps({"code": js_text, "options": options, "globals": globals}),
        timeout=eval_timeout,
        context=jshint_functions_text,
        container_engine=container_engine,
    )

    def dump_jshint_error() -> None:
        raise RuntimeError(
            'jshint failed to run successfully\nreturncode: %d\nstdout: "%s"\nstderr: "%s"'
            % (returncode, stdout, stderr)
        )

    if returncode == -1:
        _logger.warning("jshint process timed out")

    if returncode != 0:
        dump_jshint_error()

    try:
        jshint_json = json.loads(stdout)
    except ValueError:
        dump_jshint_error()

    jshint_errors: list[str] = []

    js_text_lines = js_text.split("\n")

    for jshint_error_obj in jshint_json.get("errors", []):
        text = "JSHINT: " + js_text_lines[jshint_error_obj["line"] - 1] + "\n"
        text += "JSHINT: " + " " * (jshint_error_obj["character"] - 1) + "^\n"
        text += "JSHINT: {}: {}".format(
            jshint_error_obj["code"],
            jshint_error_obj["reason"],
        )
        jshint_errors.append(text)

    return JSHintJSReturn(jshint_errors, jshint_json.get("globals", []))


def print_js_hint_messages(js_hint_messages: list[str], source_line: SourceLine | None) -> None:
    """Log the message from JSHint, using the line number."""
    if source_line is not None:
        for js_hint_message in js_hint_messages:
            _logger.warning(source_line.makeError(js_hint_message))


def validate_js_expressions(
    tool: CommentedMap,
    schema: Schema,
    jshint_options: dict[str, list[str] | str | int] | None = None,
    container_engine: str = "docker",
    eval_timeout: float = 60,
) -> None:
    if tool.get("requirements") is None:
        return
    debug = _logger.isEnabledFor(logging.DEBUG)
    requirements = tool["requirements"]

    default_globals = ["self", "inputs", "runtime", "console"]

    for prop in reversed(requirements):
        if prop["class"] == "InlineJavascriptRequirement":
            expression_lib = prop.get("expressionLib", [])
            break
    else:
        return

    js_globals = copy.deepcopy(default_globals)

    for i, expression_lib_line in enumerate(expression_lib):
        expression_lib_line_errors, expression_lib_line_globals = jshint_js(
            expression_lib_line,
            js_globals,
            jshint_options,
            container_engine,
            eval_timeout,
        )
        js_globals.extend(expression_lib_line_globals)
        print_js_hint_messages(
            expression_lib_line_errors,
            SourceLine(expression_lib, i, include_traceback=debug),
        )

    expressions = get_expressions(tool, schema)

    for expression, source_line in expressions:
        unscanned_str = expression.strip()
        try:
            scan_slice = scan_expression(unscanned_str)
        except SubstitutionError as se:
            if source_line:
                source_line.raise_type = WorkflowException
                raise source_line.makeError(str(se)) from se
            else:
                raise se

        while scan_slice:
            if unscanned_str[scan_slice[0]] == "$":
                code_fragment = unscanned_str[scan_slice[0] + 1 : scan_slice[1]]
                code_fragment_js = code_fragment_to_js(code_fragment, "")
                expression_errors, _ = jshint_js(
                    code_fragment_js,
                    js_globals,
                    jshint_options,
                    container_engine,
                    eval_timeout,
                )
                print_js_hint_messages(expression_errors, source_line)

            unscanned_str = unscanned_str[scan_slice[1] :]
            scan_slice = scan_expression(unscanned_str)



import logging
import subprocess
import sys
from collections.abc import MutableMapping
from io import StringIO
from pathlib import Path
from typing import cast

from cwltool.context import RuntimeContext
from cwltool.factory import Factory
from cwltool.utils import CWLObjectType
from cwltool.workflow_job import WorkflowJobStep

from .util import get_data, needs_docker


def test_replace_default_stdout_stderr() -> None:
    """Test our ability to replace the default stdout/err."""

    # break stdout & stderr
    original_stdout = sys.stdout
    original_stderr = sys.stderr

    sys.stdout = ""
    sys.stderr = ""

    runtime_context = RuntimeContext()
    runtime_context.default_stdout = subprocess.DEVNULL  # type: ignore
    runtime_context.default_stderr = subprocess.DEVNULL  # type: ignore
    factory = Factory(None, None, runtime_context)
    echo = factory.make(get_data("tests/echo.cwl"))

    assert echo(inp="foo") == {"out": "foo\n"}
    sys.stdout = original_stdout
    sys.stderr = original_stderr


@needs_docker
def test_workflow_job_step_name_callback() -> None:
    """Test ability to hook custom workflow step naming"""

    stream = StringIO()
    streamhandler = logging.StreamHandler(stream)
    _logger = logging.getLogger("cwltool")
    _logger.addHandler(streamhandler)

    try:
        runtime_context = RuntimeContext()

        def step_name_hook(step: WorkflowJobStep, job: CWLObjectType) -> str:
            j1 = cast(MutableMapping[str, CWLObjectType], job)
            inp = cast(MutableMapping[str, str], j1.get("revtool_input", j1.get("sorted_input")))
            return "{} on {}".format(
                step.name,
                inp.get("basename"),
            )

        runtime_context.workflow_job_step_name_callback = step_name_hook

        factory = Factory(None, None, runtime_context)
        revsort = factory.make(get_data("tests/wf/revsort.cwl"))

        result = revsort(
            workflow_input={
                "class": "File",
                "location": Path(get_data("tests/wf/whale.txt")).as_uri(),
                "format": "https://www.iana.org/assignments/media-types/text/plain",
            }
        )

        result = cast(CWLObjectType, result)

        sorted_out = cast(MutableMapping[str, str], result["sorted_output"])
        loc = sorted_out["location"]

        assert result == {
            "sorted_output": {
                "basename": "output.txt",
                "checksum": "sha1$b9214658cc453331b62c2282b772a5c063dbd284",
                "class": "File",
                "http://commonwl.org/cwltool#generation": 0,
                "nameext": ".txt",
                "nameroot": "output",
                "size": 1111,
                "location": loc,
            },
        }

        assert "rev on whale.txt" in stream.getvalue()
    finally:
        _logger.removeHandler(streamhandler)



"""Tests for --make-template."""

from schema_salad.sourceline import cmap

from cwltool import main


def test_anonymous_record() -> None:
    inputs = cmap([{"type": "record", "fields": []}])
    assert main.generate_example_input(inputs, None) == ({}, "Anonymous record type.")


def test_union() -> None:
    """Test for --make-template for a union type."""
    inputs = cmap(["string", "string[]"])
    assert main.generate_example_input(inputs, None) == (
        "a_string",
        "one of type 'string', type 'string[]'",
    )


def test_optional_union() -> None:
    """Test for --make-template for an optional union type."""
    inputs = cmap(["null", "string", "string[]"])
    assert main.generate_example_input(inputs, None) == (
        "a_string",
        "one of type 'string', type 'string[]' (optional)",
    )



import pytest
from cwl_utils.sandboxjs import code_fragment_to_js
from schema_salad.avro.schema import Names
from schema_salad.utils import yaml_no_ts

from cwltool import process, validate_js

TEST_CWL = """
cwlVersion: v1.0
class: CommandLineTool
baseCommand: echo

requirements:
  - class: InlineJavascriptRequirement
inputs:
  - id: parameter
    inputBinding:
      valueFrom: string before $(kjdbfkjd) string after
    type: string?

outputs: []
"""


def test_get_expressions() -> None:
    yaml = yaml_no_ts()
    test_cwl_yaml = yaml.load(TEST_CWL)
    schema = process.get_schema("v1.0")[1]
    assert isinstance(schema, Names)
    clt_schema = schema.names["org.w3id.cwl.cwl.CommandLineTool"]

    exprs = validate_js.get_expressions(test_cwl_yaml, clt_schema)

    assert len(exprs) == 1


def test_validate_js_expressions(caplog: pytest.LogCaptureFixture) -> None:
    """Test invalid JS expression."""
    yaml = yaml_no_ts()
    test_cwl_yaml = yaml.load(TEST_CWL)
    schema = process.get_schema("v1.0")[1]
    assert isinstance(schema, Names)
    clt_schema = schema.names["org.w3id.cwl.cwl.CommandLineTool"]

    validate_js.validate_js_expressions(test_cwl_yaml, clt_schema)

    assert (
        " JSHINT: (function(){return ((kjdbfkjd));})()\n"
        " JSHINT:                      ^\n"
        " JSHINT: W117: 'kjdbfkjd' is not defined."
    ) in caplog.text


def test_js_hint_basic() -> None:
    result = validate_js.jshint_js(
        """
    function funcName(){
    }
    """,
        [],
    )

    assert result.errors == []
    assert result.globals == ["funcName"]


def test_js_hint_reports_invalid_js() -> None:
    assert len(validate_js.jshint_js("<INVALID JS>").errors) > 1


def test_js_hint_warn_on_es6() -> None:
    assert len(validate_js.jshint_js(code_fragment_to_js("((() => 4)())"), []).errors) == 1


def test_js_hint_error_on_undefined_name() -> None:
    assert len(validate_js.jshint_js(code_fragment_to_js("undefined_name()")).errors) == 1


def test_js_hint_set_defined_name() -> None:
    assert (
        len(validate_js.jshint_js(code_fragment_to_js("defined_name()"), ["defined_name"]).errors)
        == 0
    )



"""Common errors.

WorkflowException and GraphTargetMissingException are aliased to
equivalent errors from cwl_utils.errors and re-exported by this module
to avoid breaking the interface for other code.

"""

# flake8: noqa: F401

from cwl_utils.errors import GraphTargetMissingException as GraphTargetMissingException
from cwl_utils.errors import WorkflowException as WorkflowException

__all__ = (
    "GraphTargetMissingException",
    "WorkflowException",
    "UnsupportedRequirement",
    "ArgumentException",
)


class UnsupportedRequirement(WorkflowException):
    pass


class ArgumentException(Exception):
    """Mismatched command line arguments provided."""



"""Test that all temporary directories respect the --tmpdir-prefix and --tmp-outdir-prefix options."""

import copy
import logging
import os
import re
import shutil
import subprocess
import sys
from pathlib import Path
from typing import cast

import pytest
from ruamel.yaml.comments import CommentedMap
from schema_salad.avro import schema
from schema_salad.sourceline import cmap

from cwltool.builder import Builder
from cwltool.command_line_tool import CommandLineTool
from cwltool.context import LoadingContext, RuntimeContext
from cwltool.docker import DockerCommandLineJob
from cwltool.errors import WorkflowException
from cwltool.job import JobBase
from cwltool.main import main
from cwltool.pathmapper import MapperEnt
from cwltool.singularity import _IMAGES, _IMAGES_LOCK, SingularityCommandLineJob
from cwltool.stdfsaccess import StdFsAccess
from cwltool.update import INTERNAL_VERSION, ORIGINAL_CWLVERSION
from cwltool.utils import CWLObjectType, create_tmp_dir

from .util import get_data, get_main_output, needs_docker, needs_singularity


@pytest.fixture(autouse=True)
def clear_singularity_image_cache() -> None:
    with _IMAGES_LOCK:
        _IMAGES.clear()


def test_docker_commandLineTool_job_tmpdir_prefix(tmp_path: Path) -> None:
    """Test that docker enabled CommandLineTool respects temp directory directives."""
    loading_context = LoadingContext(
        {
            "metadata": {
                "cwlVersion": INTERNAL_VERSION,
                ORIGINAL_CWLVERSION: INTERNAL_VERSION,
            }
        }
    )
    clt = CommandLineTool(
        cast(
            CommentedMap,
            cmap(
                {
                    "cwlVersion": INTERNAL_VERSION,
                    "class": "CommandLineTool",
                    "inputs": [],
                    "outputs": [],
                    "requirements": [
                        {
                            "class": "DockerRequirement",
                            "dockerPull": "docker.io/debian:stable-slim",
                        }
                    ],
                }
            ),
        ),
        loading_context,
    )
    tmpdir_prefix = str(tmp_path / "1")
    tmp_outdir_prefix = str(tmp_path / "2")
    runtime_context = RuntimeContext(
        {
            "tmpdir_prefix": tmpdir_prefix,
            "tmp_outdir_prefix": tmp_outdir_prefix,
        }
    )
    job = next(clt.job({}, lambda output, process_status: None, runtime_context))
    assert isinstance(job, JobBase)
    assert job.stagedir and job.stagedir.startswith(tmpdir_prefix)
    assert job.tmpdir and job.tmpdir.startswith(tmpdir_prefix)
    assert job.outdir and job.outdir.startswith(tmp_outdir_prefix)


def test_commandLineTool_job_tmpdir_prefix(tmp_path: Path) -> None:
    """Test that non-docker enabled CommandLineTool respects temp directory directives."""
    loading_context = LoadingContext(
        {
            "metadata": {
                "cwlVersion": INTERNAL_VERSION,
                ORIGINAL_CWLVERSION: INTERNAL_VERSION,
            }
        }
    )
    clt = CommandLineTool(
        cast(
            CommentedMap,
            cmap(
                {
                    "cwlVersion": INTERNAL_VERSION,
                    "class": "CommandLineTool",
                    "inputs": [],
                    "outputs": [],
                    "requirements": [],
                }
            ),
        ),
        loading_context,
    )
    tmpdir_prefix = str(tmp_path / "1")
    tmp_outdir_prefix = str(tmp_path / "2")
    runtime_context = RuntimeContext(
        {
            "tmpdir_prefix": tmpdir_prefix,
            "tmp_outdir_prefix": tmp_outdir_prefix,
        }
    )
    job = next(clt.job({}, lambda output, process_status: None, runtime_context))
    assert isinstance(job, JobBase)
    assert job.stagedir and job.stagedir.startswith(tmpdir_prefix)
    assert job.tmpdir and job.tmpdir.startswith(tmpdir_prefix)
    assert job.outdir and job.outdir.startswith(tmp_outdir_prefix)


@needs_docker
def test_dockerfile_tmpdir_prefix(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    """Test that DockerCommandLineJob.get_image respects temp directory directives."""
    monkeypatch.setattr(target=subprocess, name="check_call", value=lambda *args, **kwargs: True)
    (tmp_path / "out").mkdir()
    tmp_outdir_prefix = tmp_path / "out" / "1"
    (tmp_path / "3").mkdir()
    tmpdir_prefix = str(tmp_path / "3" / "ttmp")
    runtime_context = RuntimeContext(
        {"tmpdir_prefix": tmpdir_prefix, "user_space_docker_cmd": None}
    )
    builder = Builder(
        {},
        [],
        [],
        {},
        schema.Names(),
        [],
        [],
        {},
        None,
        None,
        StdFsAccess,
        StdFsAccess(""),
        None,
        0.1,
        False,
        False,
        False,
        "no_listing",
        runtime_context.get_outdir(),
        runtime_context.get_tmpdir(),
        runtime_context.get_stagedir(),
        INTERNAL_VERSION,
        "docker",
    )
    assert DockerCommandLineJob(
        builder, {}, CommandLineTool.make_path_mapper, [], [], ""
    ).get_image(
        {
            "class": "DockerRequirement",
            "dockerFile": "FROM debian:stable-slim",
            "dockerImageId": sys._getframe().f_code.co_name,
        },
        pull_image=True,
        force_pull=True,
        tmp_outdir_prefix=str(tmp_outdir_prefix),
    )
    children = sorted(tmp_outdir_prefix.parent.glob("*"))
    assert len(children) == 1
    subdir = tmp_path / children[0]
    assert len(sorted(subdir.glob("*"))) == 1
    assert (subdir / "Dockerfile").exists()


@needs_docker
def test_dockerfile_build(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    """Test that DockerCommandLineJob.get_image builds a Dockerfile."""
    (tmp_path / "out").mkdir()
    tmp_outdir_prefix = tmp_path / "out" / "1"
    (tmp_path / "3").mkdir()
    tmpdir_prefix = str(tmp_path / "3" / "ttmp")
    runtime_context = RuntimeContext(
        {"tmpdir_prefix": tmpdir_prefix, "user_space_docker_cmd": None}
    )
    builder = Builder(
        {},
        [],
        [],
        {},
        schema.Names(),
        [],
        [],
        {},
        None,
        None,
        StdFsAccess,
        StdFsAccess(""),
        None,
        0.1,
        False,
        False,
        False,
        "no_listing",
        runtime_context.get_outdir(),
        runtime_context.get_tmpdir(),
        runtime_context.get_stagedir(),
        INTERNAL_VERSION,
        "docker",
    )

    docker_image_id = sys._getframe().f_code.co_name

    assert DockerCommandLineJob(
        builder, {}, CommandLineTool.make_path_mapper, [], [], ""
    ).get_image(
        {
            "class": "DockerRequirement",
            "dockerFile": "FROM debian:stable-slim",
            "dockerImageId": docker_image_id,
        },
        pull_image=False,
        force_pull=False,
        tmp_outdir_prefix=str(tmp_outdir_prefix),
    )
    output = subprocess.check_output(
        ["docker", "images", "--quiet", docker_image_id], stderr=subprocess.STDOUT, text=True
    )

    # If the output is empty, the image doesn't exist
    assert output.strip(), f"Docker image {docker_image_id} does not exist"


@needs_singularity
def test_dockerfile_singularity_build(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    """Test that SingularityCommandLineJob.get_image builds a Dockerfile with Singularity."""
    tmppath = Path(os.environ.get("APPTAINER_TMPDIR", tmp_path))
    # some HPC not allowed to execute on /tmp so allow user to define temp path with APPTAINER_TMPDIR
    # FATAL:   Unable to create build: 'noexec' mount option set on /tmp, temporary root filesystem
    monkeypatch.setattr(target=subprocess, name="check_call", value=lambda *args, **kwargs: True)
    (tmppath / "out").mkdir(exist_ok=True)
    tmp_outdir_prefix = tmppath / "out" / "1"
    (tmppath / "3").mkdir(exist_ok=True)
    tmpdir_prefix = str(tmppath / "3" / "ttmp")
    runtime_context = RuntimeContext(
        {"tmpdir_prefix": tmpdir_prefix, "user_space_docker_cmd": None}
    )
    builder = Builder(
        {},
        [],
        [],
        {},
        schema.Names(),
        [],
        [],
        {},
        None,
        None,
        StdFsAccess,
        StdFsAccess(""),
        None,
        0.1,
        True,
        False,
        False,
        "no_listing",
        runtime_context.get_outdir(),
        runtime_context.get_tmpdir(),
        runtime_context.get_stagedir(),
        INTERNAL_VERSION,
        "singularity",
    )

    assert SingularityCommandLineJob(
        builder, {}, CommandLineTool.make_path_mapper, [], [], ""
    ).get_image(
        {
            "class": "DockerRequirement",
            "dockerFile": "FROM debian:stable-slim",
        },
        pull_image=True,
        tmp_outdir_prefix=str(tmp_outdir_prefix),
        force_pull=True,
    )
    children = sorted(tmp_outdir_prefix.parent.glob("*"))
    subdir = tmppath / children[0]
    children = sorted(subdir.glob("*.sif"))
    image_path = children[0]
    assert image_path.exists()
    shutil.rmtree(subdir)


@needs_singularity
def test_singularity_get_image_from_sandbox(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    """Test that SingularityCommandLineJob.get_image correctly handle sandbox image."""

    (tmp_path / "out").mkdir(exist_ok=True)
    tmp_outdir_prefix = tmp_path / "out"
    tmp_outdir_prefix.mkdir(exist_ok=True)
    (tmp_path / "tmp").mkdir(exist_ok=True)
    tmpdir_prefix = str(tmp_path / "tmp")
    runtime_context = RuntimeContext(
        {"tmpdir_prefix": tmpdir_prefix, "user_space_docker_cmd": None}
    )
    builder = Builder(
        {},
        [],
        [],
        {},
        schema.Names(),
        [],
        [],
        {},
        None,
        None,
        StdFsAccess,
        StdFsAccess(""),
        None,
        0.1,
        True,
        False,
        False,
        "no_listing",
        runtime_context.get_outdir(),
        runtime_context.get_tmpdir(),
        runtime_context.get_stagedir(),
        INTERNAL_VERSION,
        "singularity",
    )

    workdir = tmp_path / "working_dir"
    workdir.mkdir()
    repo_path = workdir / "container_repo"
    repo_path.mkdir()
    image_path = repo_path / "alpine"
    image_path.mkdir()

    # directory exists but is not an image
    monkeypatch.setattr(
        "cwltool.singularity._inspect_singularity_sandbox_image", lambda *args, **kwargs: False
    )
    req = {"class": "DockerRequirement", "dockerPull": f"{image_path}"}
    res = SingularityCommandLineJob(
        builder, {}, CommandLineTool.make_path_mapper, [], [], ""
    ).get_image(
        req,
        pull_image=False,
        tmp_outdir_prefix=str(tmp_outdir_prefix),
        force_pull=False,
    )
    assert req["dockerPull"] == f"{image_path}"
    assert res is False

    # directory exists and is an image:
    monkeypatch.setattr(
        "cwltool.singularity._inspect_singularity_sandbox_image", lambda *args, **kwargs: True
    )
    req = {"class": "DockerRequirement", "dockerPull": f"{image_path}"}
    res = SingularityCommandLineJob(
        builder, {}, CommandLineTool.make_path_mapper, [], [], ""
    ).get_image(
        req,
        pull_image=False,
        tmp_outdir_prefix=str(tmp_outdir_prefix),
        force_pull=False,
    )
    assert req["dockerImageId"] == str(image_path)
    assert res

    # test that dockerImageId is set and image exists:
    req = {"class": "DockerRequirement", "dockerImageId": f"{image_path}"}
    res = SingularityCommandLineJob(
        builder, {}, CommandLineTool.make_path_mapper, [], [], ""
    ).get_image(
        req,
        pull_image=False,
        tmp_outdir_prefix=str(tmp_outdir_prefix),
        force_pull=False,
    )
    assert req["dockerImageId"] == str(image_path)
    assert res


@needs_singularity
def test_singularity_image_base_path(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path, caplog: pytest.LogCaptureFixture
) -> None:
    tmp_outdir_prefix = tmp_path / "out"
    tmp_outdir_prefix.mkdir()
    tmpdir = tmp_path / "tmp"
    tmpdir.mkdir()

    workdir = tmp_path / "working_dir"
    workdir.mkdir()
    repo_path = (workdir / "container_repo").absolute()
    repo_path.mkdir()
    image_path = (repo_path / "alpine").absolute()
    image_path.mkdir()

    runtime_context = RuntimeContext(
        {"tmp_outdir_prefix": str(tmp_outdir_prefix), "tmpdir": str(tmpdir)}
    )
    builder = Builder(
        {},
        [],
        [],
        {},
        schema.Names(),
        [],
        [],
        {},
        None,
        None,
        StdFsAccess,
        StdFsAccess(""),
        None,
        0.1,
        True,
        False,
        False,
        "no_listing",
        runtime_context.get_outdir(),
        runtime_context.get_tmpdir(),
        runtime_context.get_stagedir(),
        INTERNAL_VERSION,
        "singularity",
    )

    monkeypatch.setattr(
        "cwltool.singularity._inspect_singularity_sandbox_image", lambda *args, **kwargs: True
    )

    initial_requirements = {"class": "DockerRequirement", "dockerPull": "alpine"}
    requirements1 = copy.deepcopy(initial_requirements)
    # get image from sandbox_base_path option
    res_get_image = SingularityCommandLineJob(
        builder, {}, CommandLineTool.make_path_mapper, [], [], ""
    ).get_image(
        requirements1,
        pull_image=False,
        tmp_outdir_prefix=str(tmp_outdir_prefix),
        force_pull=False,
        sandbox_base_path=str(repo_path),
    )
    assert res_get_image
    assert requirements1["dockerImageId"] == str(image_path)

    requirements2: CWLObjectType = {"class": "DockerRequirement", "dockerPull": "alpine"}
    res_get_req = SingularityCommandLineJob(
        builder, {}, CommandLineTool.make_path_mapper, [], [], ""
    ).get_from_requirements(
        requirements2,
        pull_image=False,
        force_pull=False,
        tmp_outdir_prefix=str(tmp_outdir_prefix),
        image_base_path=str(repo_path),
    )
    assert res_get_req == str(image_path)

    with _IMAGES_LOCK:
        _IMAGES.clear()
    requirements3: CWLObjectType = {"class": "DockerRequirement", "dockerPull": "alpine"}
    # get requirements from without sandbox image path
    # should return an error
    with pytest.raises(WorkflowException):
        res_get_req = SingularityCommandLineJob(
            builder, {}, CommandLineTool.make_path_mapper, [], [], ""
        ).get_from_requirements(
            requirements3,
            pull_image=False,
            force_pull=False,
            tmp_outdir_prefix=str(tmp_outdir_prefix),
        )

    # get image from CWL_SINGULARITY_IMAGES env var
    requirements4 = copy.deepcopy(initial_requirements)
    with monkeypatch.context() as m:
        m.setenv("CWL_SINGULARITY_IMAGES", str(repo_path))
        res_get_image = SingularityCommandLineJob(
            builder, {}, CommandLineTool.make_path_mapper, [], [], ""
        ).get_image(
            requirements4,
            pull_image=False,
            tmp_outdir_prefix=str(tmp_outdir_prefix),
            force_pull=False,
        )
        assert res_get_image
        assert requirements4["dockerImageId"] == str(image_path)

    with _IMAGES_LOCK:
        _IMAGES.clear()
    with caplog.at_level(logging.INFO):
        requirements5: CWLObjectType = {
            "class": "DockerRequirement",
            "dockerPull": "alpine",
            "dockerImageId": str(image_path),
        }
        res_get_req = SingularityCommandLineJob(
            builder, {}, CommandLineTool.make_path_mapper, [], [], ""
        ).get_from_requirements(
            requirements5,
            pull_image=False,
            force_pull=False,
            tmp_outdir_prefix=str(tmp_outdir_prefix),
        )
        assert res_get_req == str(image_path)
        assert (
            f"Non-portable: using an absolute file path in a 'dockerImageId': {image_path}"
            in caplog.text
        )


def test_docker_tmpdir_prefix(tmp_path: Path) -> None:
    """Test that DockerCommandLineJob respects temp directory directives."""
    (tmp_path / "3").mkdir()
    tmpdir_prefix = str(tmp_path / "3" / "ttmp")
    runtime_context = RuntimeContext(
        {"tmpdir_prefix": tmpdir_prefix, "user_space_docker_cmd": None}
    )
    builder = Builder(
        {},
        [],
        [],
        {},
        schema.Names(),
        [],
        [],
        {},
        None,
        None,
        StdFsAccess,
        StdFsAccess(""),
        None,
        0.1,
        False,
        False,
        False,
        "no_listing",
        runtime_context.get_outdir(),
        runtime_context.get_tmpdir(),
        runtime_context.get_stagedir(),
        INTERNAL_VERSION,
        "docker",
    )
    job = DockerCommandLineJob(builder, {}, CommandLineTool.make_path_mapper, [], [], "")
    runtime: list[str] = []

    volume_writable_file = MapperEnt(
        resolved=get_data("tests/2.fastq"), target="foo", type=None, staged=None
    )
    (tmp_path / "1").mkdir()
    job.add_writable_file_volume(
        runtime, volume_writable_file, None, str(tmp_path / "1" / "writable_file")
    )
    children = sorted((tmp_path / "1").glob("*"))
    assert len(children) == 1
    subdir = tmp_path / children[0]
    assert subdir.name.startswith("writable_file")
    assert len(sorted(subdir.glob("*"))) == 1
    assert (subdir / "2.fastq").exists()

    resolved_writable_dir = tmp_path / "data_orig"
    resolved_writable_dir.mkdir(parents=True)
    volume_dir = MapperEnt(
        resolved=str(resolved_writable_dir), target="bar", type=None, staged=None
    )
    (tmp_path / "2").mkdir()
    job.add_writable_directory_volume(runtime, volume_dir, None, str(tmp_path / "2" / "dir"))
    children = sorted((tmp_path / "2").glob("*"))
    assert len(children) == 1
    subdir = tmp_path / "2" / children[0]
    assert subdir.name.startswith("dir")
    assert len(sorted(subdir.glob("*"))) == 1
    assert (subdir / "data_orig").exists()

    cidfile = job.create_runtime({}, runtime_context)[1]
    assert cidfile and cidfile.startswith(tmpdir_prefix)

    volume_file = MapperEnt(resolved="Hoopla!", target="baz", type=None, staged=None)
    (tmp_path / "4").mkdir()
    job.create_file_and_add_volume(runtime, volume_file, None, None, str(tmp_path / "4" / "file"))
    children = sorted((tmp_path / "4").glob("*"))
    assert len(children) == 1
    subdir = tmp_path / "4" / children[0]
    assert subdir.name.startswith("file")
    assert len(sorted(subdir.glob("*"))) == 1
    assert (subdir / "baz").exists()


def test_runtimeContext_respects_tmpdir_prefix(tmp_path: Path) -> None:
    """Test that RuntimeContext helper methods respects tmpdir_prefix."""
    tmpdir_prefix = str(tmp_path / "foo")
    runtime_context = RuntimeContext({"tmpdir_prefix": tmpdir_prefix})
    assert runtime_context.get_tmpdir().startswith(tmpdir_prefix)
    assert runtime_context.get_stagedir().startswith(tmpdir_prefix)
    assert runtime_context.create_tmpdir().startswith(tmpdir_prefix)
    assert create_tmp_dir(tmpdir_prefix).startswith(tmpdir_prefix)


def test_runtimeContext_respects_tmp_outdir_prefix(tmp_path: Path) -> None:
    """Test that RuntimeContext helper methods respects tmp_outdir_prefix."""
    tmpdir_prefix = str(tmp_path / "foo")
    runtime_context = RuntimeContext({"tmp_outdir_prefix": tmpdir_prefix})
    assert runtime_context.get_outdir().startswith(tmpdir_prefix)
    assert runtime_context.create_outdir().startswith(tmpdir_prefix)


def test_remove_tmpdirs(tmp_path: Path) -> None:
    """Test that the tmpdirs are removed after the job execution."""
    assert (
        main(
            [
                "--tmpdir-prefix",
                str(f"{tmp_path}/"),
                get_data("tests/wf/hello_single_tool.cwl"),
                "--message",
                "Hello",
            ]
        )
        == 0
    )
    assert len(list(tmp_path.iterdir())) == 0


def test_leave_tmpdirs(tmp_path: Path) -> None:
    """Test that the tmpdirs including input staging directories are retained after the job execution."""
    error_code, stdout, stderr = get_main_output(
        [
            "--debug",
            "--tmpdir-prefix",
            str(f"{tmp_path}/tmp/"),
            "--leave-tmpdir",
            "--outdir",
            str(f"{tmp_path}/out/"),
            get_data("tests/env4.cwl"),
        ]
    )
    assert error_code == 0
    assert re.search(rf"\"{re.escape(str(tmp_path))}/tmp/.*/env0\.py\"", stderr)
    assert len(list((tmp_path / "tmp").iterdir())) == 3
    assert len(list((tmp_path / "tmp").glob("**/env0.py"))) == 1



import pathlib
from builtins import set as _set
from collections.abc import Iterable, Iterator
from typing import IO, Any, List, Optional, Tuple, Union, overload

from rdflib import query
from rdflib.collection import Collection
from rdflib.paths import Path
from rdflib.resource import Resource
from rdflib.term import BNode, Identifier, Node

class Graph(Node):
    base: Any = ...
    context_aware: bool = ...
    formula_aware: bool = ...
    default_union: bool = ...
    def __init__(
        self,
        store: str = ...,
        identifier: Any | None = ...,
        namespace_manager: Any | None = ...,
        base: Any | None = ...,
    ) -> None: ...
    store: Any = ...
    identifier: Any = ...
    namespace_manager: Any = ...
    def toPython(self) -> str: ...
    def destroy(self, configuration: Any) -> None: ...
    def commit(self) -> None: ...
    def rollback(self) -> None: ...
    def open(self, configuration: Any, create: bool = ...) -> Any: ...
    def close(self, commit_pending_transaction: bool = ...) -> None: ...
    def add(self, triple: Any) -> None: ...
    def addN(self, quads: Any) -> None: ...
    def remove(self, triple: Any) -> None: ...
    def triples(
        self,
        triple: tuple[
            str | Identifier | None,
            str | Identifier | None,
            Identifier | None,
        ],
    ) -> Iterator[tuple[Identifier, Identifier, Identifier]]: ...
    def __getitem__(
        self, item: slice | Path | Node
    ) -> Iterator[
        tuple[Identifier, Identifier, Identifier] | tuple[Identifier, identifier] | Node
    ]: ...
    def __contains__(self, triple: Any) -> bool: ...
    def __add__(self, other: Any) -> Graph: ...
    def set(self, triple: Any) -> None: ...
    def subjects(self, predicate: Any | None = ..., object: Any | None = ...) -> Iterable[Node]: ...
    def predicates(self, subject: Any | None = ..., object: Any | None = ...) -> Iterable[Node]: ...
    def objects(
        self, subject: Any | None = ..., predicate: Any | None = ...
    ) -> Iterable[Identifier]: ...
    def subject_predicates(self, object: Any | None = ...) -> None: ...
    def subject_objects(self, predicate: Any | None = ...) -> None: ...
    def predicate_objects(self, subject: Any | None = ...) -> None: ...
    def triples_choices(self, triple: Any, context: Any | None = ...) -> None: ...
    def value(
        self,
        subject: Any | None = ...,
        predicate: Any = ...,
        object: Any | None = ...,
        default: Any | None = ...,
        any: bool = ...,
    ) -> Any: ...
    def label(self, subject: Any, default: str = ...) -> Any: ...
    def preferredLabel(
        self,
        subject: Any,
        lang: Any | None = ...,
        default: Any | None = ...,
        labelProperties: Any = ...,
    ) -> list[tuple[Any, Any]]: ...
    def comment(self, subject: Any, default: str = ...) -> Any: ...
    def items(self, list: Any) -> Iterator[Any]: ...
    def transitiveClosure(self, func: Any, arg: Any, seen: Any | None = ...) -> Iterator[Any]: ...
    def transitive_objects(
        self, subject: Any, property: Any, remember: Any | None = ...
    ) -> Iterator[Any]: ...
    def transitive_subjects(
        self, predicate: Any, object: Any, remember: Any | None = ...
    ) -> Iterator[Any]: ...
    def seq(self, subject: Any) -> Seq | None: ...
    def qname(self, uri: Any) -> Any: ...
    def compute_qname(self, uri: Any, generate: bool = ...) -> Any: ...
    def bind(
        self, prefix: Any, namespace: Any, override: bool = ..., replace: bool = ...
    ) -> Any: ...
    def namespaces(self) -> Iterator[tuple[Any, Any]]: ...
    def absolutize(self, uri: Any, defrag: int = ...) -> Any: ...

    # no destination and non-None positional encoding
    @overload
    def serialize(
        self,
        destination: None,
        format: str,
        base: str | None,
        encoding: str,
        **args: Any,
    ) -> bytes: ...

    # no destination and non-None keyword encoding
    @overload
    def serialize(
        self,
        destination: None = ...,
        format: str = ...,
        base: str | None = ...,
        *,
        encoding: str,
        **args: Any,
    ) -> bytes: ...

    # no destination and None encoding
    @overload
    def serialize(
        self,
        destination: None = ...,
        format: str = ...,
        base: str | None = ...,
        encoding: None = ...,
        **args: Any,
    ) -> str: ...

    # non-None destination
    @overload
    def serialize(
        self,
        destination: str | pathlib.PurePath | IO[bytes],
        format: str = ...,
        base: str | None = ...,
        encoding: str | None = ...,
        **args: Any,
    ) -> "Graph": ...

    # fallback
    @overload
    def serialize(
        self,
        destination: str | pathlib.PurePath | IO[bytes] | None = ...,
        format: str = ...,
        base: str | None = ...,
        encoding: str | None = ...,
        **args: Any,
    ) -> Union[bytes, str, "Graph"]: ...
    def parse(
        self,
        source: Any | None = ...,
        publicID: Any | None = ...,
        format: str | None = ...,
        location: Any | None = ...,
        file: Any | None = ...,
        data: Any | None = ...,
        **args: Any,
    ) -> "Graph": ...
    def load(self, source: Any, publicID: Any | None = ..., format: str = ...) -> "Graph": ...
    def query(
        self,
        query_object: Any,
        processor: str = ...,
        result: str = ...,
        initNs: Any | None = ...,
        initBindings: Any | None = ...,
        use_store_provided: bool = ...,
        **kwargs: Any,
    ) -> query.Result: ...
    def update(
        self,
        update_object: Any,
        processor: str = ...,
        initNs: Any | None = ...,
        initBindings: Any | None = ...,
        use_store_provided: bool = ...,
        **kwargs: Any,
    ) -> Any: ...
    def n3(self) -> str: ...
    def isomorphic(self, other: Any) -> bool: ...
    def connected(self) -> bool: ...
    def all_nodes(self) -> _set[Node]: ...
    def collection(self, identifier: Any) -> Collection: ...
    def resource(self, identifier: Any) -> Resource: ...
    def skolemize(
        self,
        new_graph: Any | None = ...,
        bnode: Any | None = ...,
        authority: Any | None = ...,
        basepath: Any | None = ...,
    ) -> Graph: ...
    def de_skolemize(self, new_graph: Any | None = ..., uriref: Any | None = ...) -> Graph: ...

class ConjunctiveGraph(Graph):
    context_aware: bool = ...
    default_union: bool = ...
    default_context: Any = ...
    def __init__(
        self,
        store: str = ...,
        identifier: Any | None = ...,
        default_graph_base: Any | None = ...,
    ) -> None: ...
    def add(self, triple_or_quad: Any) -> None: ...
    def addN(self, quads: Any) -> None: ...
    def remove(self, triple_or_quad: Any) -> None: ...
    # def triples(self, triple_or_quad: Tuple[Optional[Union[str, BNode]], Optional[Union[str, BNode]], Optional[BNode]], context: Tuple[Optional[Union[str, BNode]], Optional[Union[str, BNode]], Optional[BNode]]) -> Iterator[Tuple[Identifier, Identifier, Identifier]]: ...
    def quads(self, triple_or_quad: Any | None = ...) -> None: ...
    def triples_choices(self, triple: Any, context: Any | None = ...) -> None: ...
    def contexts(self, triple: Any | None = ...) -> None: ...
    def get_context(
        self,
        identifier: Node | str | None,
        quoted: bool = ...,
        base: str | None = ...,
    ) -> Graph: ...
    def remove_context(self, context: Any) -> None: ...
    def context_id(self, uri: Any, context_id: Any | None = ...) -> Any: ...
    def parse(
        self,
        source: Any | None = ...,
        publicID: Any | None = ...,
        format: str | None = ...,
        location: Any | None = ...,
        file: Any | None = ...,
        data: Any | None = ...,
        **args: Any,
    ) -> Graph: ...

class Seq:
    def __init__(self, graph: Graph, subject: Any) -> None: ...
    def toPython(self) -> Seq: ...



"""Test singularity{,-ce} & apptainer versions."""

import pytest
from packaging.version import Version

import cwltool.singularity
from cwltool.singularity import (
    get_version,
    is_apptainer_1_or_newer,
    is_version_2_6,
    is_version_3_1_or_newer,
    is_version_3_4_or_newer,
    is_version_3_or_newer,
)


def reset_singularity_version_cache(monkeypatch: pytest.MonkeyPatch) -> None:
    """Reset the cache for testing."""
    monkeypatch.setattr(cwltool.singularity, "_SINGULARITY_VERSION", None)
    monkeypatch.setattr(cwltool.singularity, "_SINGULARITY_FLAVOR", "")


def dummy_check_output(monkeypatch: pytest.MonkeyPatch, name: str, version: str) -> None:
    monkeypatch.setattr(
        cwltool.singularity, "check_output", (lambda c, text: name + " version " + version)
    )


def test_get_version(monkeypatch: pytest.MonkeyPatch) -> None:
    """Confirm expected types of singularity.get_version()."""
    with monkeypatch.context() as m:
        dummy_check_output(m, "apptainer", "1.0.1")
        reset_singularity_version_cache(m)
        v = get_version()
        assert isinstance(v, tuple)
        assert isinstance(v[0], Version)
        assert isinstance(v[1], str)
        assert (
            cwltool.singularity._SINGULARITY_VERSION is not None
        )  # pylint: disable=protected-access
        assert len(cwltool.singularity._SINGULARITY_FLAVOR) > 0  # pylint: disable=protected-access
        v_cached = get_version()
        assert v == v_cached
        assert v[0] == Version("1.0.1")
        assert v[1] == "apptainer"

    with monkeypatch.context() as m:
        dummy_check_output(m, "singularity", "3.8.5")
        reset_singularity_version_cache(m)
        v = get_version()
        assert v[0] == Version("3.8.5")
        assert v[1] == "singularity"


def test_version_checks(monkeypatch: pytest.MonkeyPatch) -> None:
    """Confirm logic in the various singularity version checks."""
    with monkeypatch.context() as m:
        dummy_check_output(m, "apptainer", "1.0.1")
        reset_singularity_version_cache(m)
        assert is_apptainer_1_or_newer()
        assert not is_version_2_6()
        assert is_version_3_or_newer()
        assert is_version_3_1_or_newer()
        assert is_version_3_4_or_newer()

    with monkeypatch.context() as m:
        dummy_check_output(m, "apptainer", "0.0.1")
        reset_singularity_version_cache(m)
        assert not is_apptainer_1_or_newer()
        assert not is_version_2_6()
        assert not is_version_3_or_newer()
        assert not is_version_3_1_or_newer()
        assert not is_version_3_4_or_newer()

    with monkeypatch.context() as m:
        dummy_check_output(m, "singularity", "0.0.1")
        reset_singularity_version_cache(m)
        assert not is_apptainer_1_or_newer()
        assert not is_version_2_6()
        assert not is_version_3_or_newer()
        assert not is_version_3_1_or_newer()
        assert not is_version_3_4_or_newer()

    with monkeypatch.context() as m:
        dummy_check_output(m, "singularity", "0.1")
        reset_singularity_version_cache(m)
        assert not is_apptainer_1_or_newer()
        assert not is_version_2_6()
        assert not is_version_3_or_newer()
        assert not is_version_3_1_or_newer()
        assert not is_version_3_4_or_newer()

    with monkeypatch.context() as m:
        dummy_check_output(m, "singularity", "2.6")
        reset_singularity_version_cache(m)
        assert not is_apptainer_1_or_newer()
        assert is_version_2_6()
        assert not is_version_3_or_newer()
        assert not is_version_3_1_or_newer()
        assert not is_version_3_4_or_newer()

    with monkeypatch.context() as m:
        dummy_check_output(m, "singularity", "3.0")
        reset_singularity_version_cache(m)
        assert not is_apptainer_1_or_newer()
        assert not is_version_2_6()
        assert is_version_3_or_newer()
        assert not is_version_3_1_or_newer()
        assert not is_version_3_4_or_newer()

    with monkeypatch.context() as m:
        dummy_check_output(m, "singularity", "3.1")
        reset_singularity_version_cache(m)
        assert not is_apptainer_1_or_newer()
        assert not is_version_2_6()
        assert is_version_3_or_newer()
        assert is_version_3_1_or_newer()
        assert not is_version_3_4_or_newer()

    with monkeypatch.context() as m:
        dummy_check_output(m, "singularity", "3.4")
        reset_singularity_version_cache(m)
        assert not is_apptainer_1_or_newer()
        assert not is_version_2_6()
        assert is_version_3_or_newer()
        assert is_version_3_1_or_newer()
        assert is_version_3_4_or_newer()

    with monkeypatch.context() as m:
        dummy_check_output(m, "singularity", "3.6.3")
        reset_singularity_version_cache(m)
        assert not is_apptainer_1_or_newer()
        assert not is_version_2_6()
        assert is_version_3_or_newer()
        assert is_version_3_1_or_newer()
        assert is_version_3_4_or_newer()



from typing import Optional

class Recipe:
    cmd: str | None
    comments: list[str]
    entrypoint: str | None
    environ: list[str]
    files: list[str]
    layer_files: dict[str, str]
    install: list[str]
    labels: list[str]
    ports: list[str]
    test: str | None
    volumes: list[str]
    workdir: str | None
    layer: int
    fromHeader: str | None
    source: Recipe | None
    def __init__(self, recipe: Recipe | None = ..., layer: int = ...) -> None: ...



"""Enables Docker software containers via the {u,}docker or podman runtimes."""

import csv
import datetime
import json
import math
import os
import shutil
import subprocess  # nosec
import sys
import threading
from collections.abc import Callable, MutableMapping
from io import StringIO  # pylint: disable=redefined-builtin
from typing import Optional, cast

import requests

from .builder import Builder
from .context import RuntimeContext
from .docker_id import docker_vm_id
from .errors import WorkflowException
from .job import ContainerCommandLineJob
from .loghandler import _logger
from .pathmapper import MapperEnt, PathMapper
from .utils import CWLObjectType, create_tmp_dir, ensure_writable

_IMAGES: set[str] = set()
_IMAGES_LOCK = threading.Lock()
__docker_machine_mounts: list[str] | None = None
__docker_machine_mounts_lock = threading.Lock()


def _get_docker_machine_mounts() -> list[str]:
    global __docker_machine_mounts
    if __docker_machine_mounts is None:
        with __docker_machine_mounts_lock:
            if "DOCKER_MACHINE_NAME" not in os.environ:
                __docker_machine_mounts = []
            else:
                __docker_machine_mounts = [
                    "/" + line.split(None, 1)[0]
                    for line in subprocess.check_output(  # nosec
                        [
                            "docker-machine",
                            "ssh",
                            os.environ["DOCKER_MACHINE_NAME"],
                            "mount",
                            "-t",
                            "vboxsf",
                        ],
                        text=True,
                    ).splitlines()
                ]
    return __docker_machine_mounts


def _check_docker_machine_path(path: str | None) -> None:
    if path is None:
        return
    mounts = _get_docker_machine_mounts()

    found = False
    for mount in mounts:
        if path.startswith(mount):
            found = True
            break

    if not found and mounts:
        name = os.environ.get("DOCKER_MACHINE_NAME", "???")
        raise WorkflowException(
            "Input path {path} is not in the list of host paths mounted "
            "into the Docker virtual machine named {name}. Already mounted "
            "paths: {mounts}.\n"
            "See https://docs.docker.com/toolbox/toolbox_install_windows/"
            "#optional-add-shared-directories for instructions on how to "
            "add this path to your VM.".format(path=path, name=name, mounts=mounts)
        )


class DockerCommandLineJob(ContainerCommandLineJob):
    """Runs a :py:class:`~cwltool.job.CommandLineJob` in a software container using the Docker engine."""

    def __init__(
        self,
        builder: Builder,
        joborder: CWLObjectType,
        make_path_mapper: Callable[[list[CWLObjectType], str, RuntimeContext, bool], PathMapper],
        requirements: list[CWLObjectType],
        hints: list[CWLObjectType],
        name: str,
    ) -> None:
        """Initialize a command line builder using the Docker software container engine."""
        super().__init__(builder, joborder, make_path_mapper, requirements, hints, name)
        self.docker_exec = "docker"

    def get_image(
        self,
        docker_requirement: dict[str, str],
        pull_image: bool,
        force_pull: bool,
        tmp_outdir_prefix: str,
    ) -> bool:
        """
        Retrieve the relevant Docker container image.

        :returns: True upon success
        """
        found = False

        if "dockerImageId" not in docker_requirement and "dockerPull" in docker_requirement:
            docker_requirement["dockerImageId"] = docker_requirement["dockerPull"]

        with _IMAGES_LOCK:
            if docker_requirement["dockerImageId"] in _IMAGES:
                return True

        if (docker_image_id := docker_requirement.get("dockerImageId")) is not None:
            try:
                manifest = json.loads(
                    str(
                        subprocess.check_output(
                            [self.docker_exec, "inspect", docker_image_id]
                        ),  # nosec
                        "utf-8",
                    )
                )
                found = manifest is not None
            except (OSError, subprocess.CalledProcessError, UnicodeError):
                pass

        cmd: list[str] = []
        if "dockerFile" in docker_requirement:
            dockerfile_dir = create_tmp_dir(tmp_outdir_prefix)
            with open(os.path.join(dockerfile_dir, "Dockerfile"), "w") as dfile:
                dfile.write(docker_requirement["dockerFile"])
            cmd = [
                self.docker_exec,
                "build",
                "--tag=%s" % str(docker_requirement["dockerImageId"]),
                dockerfile_dir,
            ]
            _logger.info(str(cmd))
            subprocess.check_call(cmd, stdout=sys.stderr)  # nosec
            found = True

        if (force_pull or not found) and pull_image:
            if "dockerPull" in docker_requirement:
                cmd = [self.docker_exec, "pull", str(docker_requirement["dockerPull"])]
                _logger.info(str(cmd))
                subprocess.check_call(cmd, stdout=sys.stderr)  # nosec
                found = True
            elif "dockerLoad" in docker_requirement:
                cmd = [self.docker_exec, "load"]
                _logger.info(str(cmd))
                if os.path.exists(docker_requirement["dockerLoad"]):
                    _logger.info(
                        "Loading docker image from %s",
                        docker_requirement["dockerLoad"],
                    )
                    with open(docker_requirement["dockerLoad"], "rb") as dload:
                        loadproc = subprocess.Popen(cmd, stdin=dload, stdout=sys.stderr)  # nosec
                else:
                    loadproc = subprocess.Popen(  # nosec
                        cmd, stdin=subprocess.PIPE, stdout=sys.stderr
                    )
                    assert loadproc.stdin is not None  # nosec
                    _logger.info("Sending GET request to %s", docker_requirement["dockerLoad"])
                    req = requests.get(docker_requirement["dockerLoad"], stream=True, timeout=60)
                    size = 0
                    for chunk in req.iter_content(1024 * 1024):
                        size += len(chunk)
                        _logger.info("\r%i bytes", size)
                        loadproc.stdin.write(chunk)
                    loadproc.stdin.close()
                rcode = loadproc.wait()
                if rcode != 0:
                    raise WorkflowException(
                        "Docker load returned non-zero exit status %i" % (rcode)
                    )
                found = True
            elif "dockerImport" in docker_requirement:
                cmd = [
                    self.docker_exec,
                    "import",
                    str(docker_requirement["dockerImport"]),
                    str(docker_requirement["dockerImageId"]),
                ]
                _logger.info(str(cmd))
                subprocess.check_call(cmd, stdout=sys.stderr)  # nosec
                found = True

        if found:
            with _IMAGES_LOCK:
                _IMAGES.add(docker_requirement["dockerImageId"])

        return found

    def get_from_requirements(
        self,
        r: CWLObjectType,
        pull_image: bool,
        force_pull: bool,
        tmp_outdir_prefix: str,
        image_base_path: str | None = None,
    ) -> str | None:
        if not shutil.which(self.docker_exec):
            raise WorkflowException(f"{self.docker_exec} executable is not available")

        if self.get_image(cast(dict[str, str], r), pull_image, force_pull, tmp_outdir_prefix):
            return cast(Optional[str], r["dockerImageId"])
        raise WorkflowException("Docker image %s not found" % r["dockerImageId"])

    @staticmethod
    def append_volume(
        runtime: list[str],
        source: str,
        target: str,
        writable: bool = False,
        skip_mkdirs: bool = False,
    ) -> None:
        """Add binding arguments to the runtime list."""
        options = [
            "type=bind",
            "source=" + source,
            "target=" + target,
        ]
        if not writable:
            options.append("readonly")
        output = StringIO()
        csv.writer(output).writerow(options)
        mount_arg = output.getvalue().strip()
        runtime.append(f"--mount={mount_arg}")
        # Unlike "--volume", "--mount" will fail if the volume doesn't already exist.
        if (not skip_mkdirs) and (not os.path.exists(source)):
            os.makedirs(source)

    def add_file_or_directory_volume(
        self, runtime: list[str], volume: MapperEnt, host_outdir_tgt: str | None
    ) -> None:
        """Append volume a file/dir mapping to the runtime option list."""
        if not volume.resolved.startswith("_:"):
            _check_docker_machine_path(volume.resolved)
            self.append_volume(runtime, volume.resolved, volume.target)

    def add_writable_file_volume(
        self,
        runtime: list[str],
        volume: MapperEnt,
        host_outdir_tgt: str | None,
        tmpdir_prefix: str,
    ) -> None:
        """Append a writable file mapping to the runtime option list."""
        if self.inplace_update:
            self.append_volume(runtime, volume.resolved, volume.target, writable=True)
        else:
            if host_outdir_tgt:
                # shortcut, just copy to the output directory
                # which is already going to be mounted
                if not os.path.exists(os.path.dirname(host_outdir_tgt)):
                    os.makedirs(os.path.dirname(host_outdir_tgt))
                shutil.copy(volume.resolved, host_outdir_tgt)
            else:
                tmpdir = create_tmp_dir(tmpdir_prefix)
                file_copy = os.path.join(tmpdir, os.path.basename(volume.resolved))
                shutil.copy(volume.resolved, file_copy)
                self.append_volume(runtime, file_copy, volume.target, writable=True)
            ensure_writable(host_outdir_tgt or file_copy)

    def add_writable_directory_volume(
        self,
        runtime: list[str],
        volume: MapperEnt,
        host_outdir_tgt: str | None,
        tmpdir_prefix: str,
    ) -> None:
        """Append a writable directory mapping to the runtime option list."""
        if volume.resolved.startswith("_:"):
            # Synthetic directory that needs creating first
            if not host_outdir_tgt:
                new_dir = os.path.join(
                    create_tmp_dir(tmpdir_prefix),
                    os.path.basename(volume.target),
                )
                self.append_volume(runtime, new_dir, volume.target, writable=True)
            elif not os.path.exists(host_outdir_tgt):
                os.makedirs(host_outdir_tgt)
        else:
            if self.inplace_update:
                self.append_volume(runtime, volume.resolved, volume.target, writable=True)
            else:
                if not host_outdir_tgt:
                    tmpdir = create_tmp_dir(tmpdir_prefix)
                    new_dir = os.path.join(tmpdir, os.path.basename(volume.resolved))
                    shutil.copytree(volume.resolved, new_dir)
                    self.append_volume(runtime, new_dir, volume.target, writable=True)
                else:
                    shutil.copytree(volume.resolved, host_outdir_tgt)
                ensure_writable(host_outdir_tgt or new_dir)

    def _required_env(self) -> dict[str, str]:
        # spec currently says "HOME must be set to the designated output
        # directory." but spec might change to designated temp directory.
        # runtime.append("--env=HOME=/tmp")
        return {
            "TMPDIR": self.CONTAINER_TMPDIR,
            "HOME": self.builder.outdir,
        }

    def create_runtime(
        self, env: MutableMapping[str, str], runtimeContext: RuntimeContext
    ) -> tuple[list[str], str | None]:
        any_path_okay = self.builder.get_requirement("DockerRequirement")[1] or False
        user_space_docker_cmd = runtimeContext.user_space_docker_cmd
        if user_space_docker_cmd:
            if "udocker" in user_space_docker_cmd:
                if runtimeContext.debug:
                    runtime = [user_space_docker_cmd, "run", "--nobanner"]
                else:
                    runtime = [user_space_docker_cmd, "--quiet", "run", "--nobanner"]
            else:
                runtime = [user_space_docker_cmd, "run"]
        else:
            runtime = [self.docker_exec, "run", "-i"]
        if runtimeContext.podman:
            runtime.append("--userns=keep-id")
        self.append_volume(
            runtime, os.path.realpath(self.outdir), self.builder.outdir, writable=True
        )
        self.append_volume(
            runtime, os.path.realpath(self.tmpdir), self.CONTAINER_TMPDIR, writable=True
        )
        self.add_volumes(
            self.pathmapper,
            runtime,
            any_path_okay=True,
            secret_store=runtimeContext.secret_store,
            tmpdir_prefix=runtimeContext.tmpdir_prefix,
        )
        if self.generatemapper is not None:
            self.add_volumes(
                self.generatemapper,
                runtime,
                any_path_okay=any_path_okay,
                secret_store=runtimeContext.secret_store,
                tmpdir_prefix=runtimeContext.tmpdir_prefix,
            )

        if user_space_docker_cmd:
            runtime = [x.replace(":ro", "") for x in runtime]
            runtime = [x.replace(":rw", "") for x in runtime]

        runtime.append("--workdir=%s" % (self.builder.outdir))
        if not user_space_docker_cmd:
            if not runtimeContext.no_read_only:
                runtime.append("--read-only=true")

            if self.networkaccess:
                if runtimeContext.custom_net:
                    runtime.append(f"--net={runtimeContext.custom_net}")
            else:
                runtime.append("--net=none")

            if self.stdout is not None:
                runtime.append("--log-driver=none")

            euid, egid = docker_vm_id()
            euid = euid if euid is not None else os.geteuid()
            egid = egid if egid is not None else os.getgid()

            if runtimeContext.no_match_user is False and (euid is not None and egid is not None):
                runtime.append("--user=%d:%d" % (euid, egid))

        if runtimeContext.rm_container:
            runtime.append("--rm")

        if self.builder.resources.get("cudaDeviceCount"):
            runtime.append("--gpus=" + str(self.builder.resources["cudaDeviceCount"]))

        cidfile_path: str | None = None
        # add parameters to docker to write a container ID file
        if runtimeContext.user_space_docker_cmd is None:
            if runtimeContext.cidfile_dir:
                cidfile_dir = runtimeContext.cidfile_dir
                if not os.path.exists(str(cidfile_dir)):
                    _logger.error(
                        "--cidfile-dir %s error:\n%s",
                        cidfile_dir,
                        "directory doesn't exist, please create it first",
                    )
                    exit(2)
                if not os.path.isdir(cidfile_dir):
                    _logger.error(
                        "--cidfile-dir %s error:\n%s",
                        cidfile_dir,
                        cidfile_dir + " is not a directory, please check it first",
                    )
                    exit(2)
            else:
                cidfile_dir = runtimeContext.create_tmpdir()

            cidfile_name = datetime.datetime.now().strftime("%Y%m%d%H%M%S-%f") + ".cid"
            if runtimeContext.cidfile_prefix is not None:
                cidfile_name = str(runtimeContext.cidfile_prefix + "-" + cidfile_name)
            cidfile_path = os.path.join(cidfile_dir, cidfile_name)
            runtime.append("--cidfile=%s" % cidfile_path)
        for key, value in self.environment.items():
            runtime.append(f"--env={key}={value}")

        res_req, _ = self.builder.get_requirement("ResourceRequirement")

        if runtimeContext.strict_memory_limit and not user_space_docker_cmd:
            ram = self.builder.resources["ram"]
            runtime.append("--memory=%dm" % ram)
        elif not user_space_docker_cmd:
            if res_req and ("ramMin" in res_req or "ramMax" in res_req):
                _logger.warning(
                    "[job %s] Skipping Docker software container '--memory' limit "
                    "despite presence of ResourceRequirement with ramMin "
                    "and/or ramMax setting. Consider running with "
                    "--strict-memory-limit for increased portability "
                    "assurance.",
                    self.name,
                )
        if runtimeContext.strict_cpu_limit and not user_space_docker_cmd:
            cpus = math.ceil(self.builder.resources["cores"])
            runtime.append(f"--cpus={cpus}")
        elif not user_space_docker_cmd:
            if res_req and ("coresMin" in res_req or "coresMax" in res_req):
                _logger.warning(
                    "[job %s] Skipping Docker software container '--cpus' limit "
                    "despite presence of ResourceRequirement with coresMin "
                    "and/or coresMax setting. Consider running with "
                    "--strict-cpu-limit for increased portability "
                    "assurance.",
                    self.name,
                )
        shm_size_od, shm_bool = self.builder.get_requirement("http://commonwl.org/cwltool#ShmSize")
        if shm_bool:
            shm_size = cast(CWLObjectType, shm_size_od)["shmSize"]
            runtime.append(f"--shm-size={shm_size}")
        return runtime, cidfile_path


class PodmanCommandLineJob(DockerCommandLineJob):
    """Runs a :py:class:`~cwltool.job.CommandLineJob` in a software container using the podman engine."""

    def __init__(
        self,
        builder: Builder,
        joborder: CWLObjectType,
        make_path_mapper: Callable[[list[CWLObjectType], str, RuntimeContext, bool], PathMapper],
        requirements: list[CWLObjectType],
        hints: list[CWLObjectType],
        name: str,
    ) -> None:
        """Initialize a command line builder using the Podman software container engine."""
        super().__init__(builder, joborder, make_path_mapper, requirements, hints, name)
        self.docker_exec = "podman"



from collections.abc import Iterable, Iterator
from typing import Any, Tuple

from _typeshed import Incomplete
from rdflib.graph import Graph, Seq
from rdflib.term import Node

class Resource:
    def __init__(self, graph: Graph, subject: Node) -> None: ...
    graph: Incomplete
    identifier: Incomplete
    def add(self, p: Node, o: Node) -> None: ...
    def remove(self, p: Node, o: Node | None = ...) -> None: ...
    def set(self, p: Node, o: Node) -> None: ...
    def subjects(self, predicate: Any | None = ...) -> Iterable[Node]: ...
    def predicates(self, o: Incomplete | None = ...) -> Iterable[Node]: ...
    def objects(self, predicate: Any | None = ...) -> Iterable[Node]: ...
    def subject_predicates(self) -> Iterator[tuple[Node, Node]]: ...
    def subject_objects(self) -> Iterator[tuple[Node, Node]]: ...
    def predicate_objects(self) -> Iterator[tuple[Node, Node]]: ...
    def value(
        self, p: Node, o: Node | None = ..., default: Any | None = ..., any: bool = ...
    ) -> Any: ...
    def label(self) -> Any: ...
    def comment(self) -> Any: ...
    def items(self) -> Iterator[Any]: ...
    def transitive_objects(self, predicate: Node, remember: Any | None = ...) -> Iterator[Any]: ...
    def transitive_subjects(self, predicate: Node, remember: Any | None = ...) -> Iterator[Any]: ...
    def seq(self) -> Seq | None: ...
    def qname(self) -> Any: ...



import os
from pathlib import Path
from typing import Any
from urllib.parse import urljoin, urlsplit

import pytest
import requests
from schema_salad.fetcher import Fetcher
from schema_salad.utils import CacheType

from cwltool.context import LoadingContext
from cwltool.load_tool import load_tool
from cwltool.main import main
from cwltool.resolver import resolve_local
from cwltool.workflow import default_make_tool

from .util import get_data, working_directory


class CWLTestFetcher(Fetcher):
    def __init__(
        self,
        cache: CacheType,
        session: requests.sessions.Session | None,
    ) -> None:
        """Create a Fetcher that provides a fixed result for testing purposes."""

    def fetch_text(self, url: str, content_types: list[str] | None = None) -> str:
        if url == "baz:bar/foo.cwl":
            return """
cwlVersion: v1.0
class: CommandLineTool
baseCommand: echo
inputs: []
outputs: []
"""
        raise RuntimeError("Not foo.cwl, was %s" % url)

    def check_exists(self, url: str) -> bool:
        return url == "baz:bar/foo.cwl"

    def urljoin(self, base: str, url: str) -> str:
        urlsp = urlsplit(url)
        if urlsp.scheme:
            return url
        basesp = urlsplit(base)

        if basesp.scheme == "keep":
            return base + "/" + url
        return urljoin(base, url)


def test_fetcher() -> None:
    def test_resolver(d: Any, a: str) -> str:
        if a.startswith("baz:bar/"):
            return a
        return "baz:bar/" + a

    loadingContext = LoadingContext(
        {
            "construct_tool_object": default_make_tool,
            "resolver": test_resolver,
            "fetcher_constructor": CWLTestFetcher,
        }
    )

    load_tool("foo.cwl", loadingContext)

    assert main(["--print-pre", "--debug", "foo.cwl"], loadingContext=loadingContext) == 0


root = Path(os.path.join(get_data("")))

path_fragments = [
    (os.path.join("tests", "echo.cwl"), "/tests/echo.cwl"),
    (os.path.join("tests", "echo.cwl") + "#main", "/tests/echo.cwl#main"),
    (str(root / "tests" / "echo.cwl"), "/tests/echo.cwl"),
    (str(root / "tests" / "echo.cwl") + "#main", "/tests/echo.cwl#main"),
]


@pytest.mark.parametrize("path,expected_path", path_fragments)
def test_resolve_local(path: str, expected_path: str) -> None:
    with working_directory(root):
        expected = root.as_uri() + expected_path
        resolved = resolve_local(None, path)
        assert resolved
        assert resolved == expected



"""Support utilities for CUDA."""

import subprocess  # nosec
import xml.dom.minidom  # nosec

from .loghandler import _logger
from .utils import CWLObjectType


def cuda_version_and_device_count() -> tuple[str, int]:
    """Determine the CUDA version and number of attached CUDA GPUs."""
    try:
        out: str | bytes = subprocess.check_output(["nvidia-smi", "-q", "-x"])  # nosec
    except Exception as e:
        _logger.debug("Error checking CUDA version with nvidia-smi: %s", e, exc_info=e)
        return ("", 0)
    dm = xml.dom.minidom.parseString(out)  # nosec

    ag = dm.getElementsByTagName("attached_gpus")
    if len(ag) < 1 or ag[0].firstChild is None:
        _logger.warning(
            "Error checking CUDA version with nvidia-smi. Missing 'attached_gpus' or it is empty.: %s",
            out,
        )
        return ("", 0)
    ag_element = ag[0].firstChild

    cv = dm.getElementsByTagName("cuda_version")
    if len(cv) < 1 or cv[0].firstChild is None:
        _logger.warning(
            "Error checking CUDA version with nvidia-smi. Missing 'cuda_version' or it is empty.: %s",
            out,
        )
        return ("", 0)
    cv_element = cv[0].firstChild

    if isinstance(cv_element, xml.dom.minidom.Text) and isinstance(
        ag_element, xml.dom.minidom.Text
    ):
        return (cv_element.data, int(ag_element.data))
    _logger.warning(
        "Error checking CUDA version with nvidia-smi. "
        "Either 'attached_gpus' or 'cuda_version' was not a text node: %s",
        out,
    )
    return ("", 0)


def cuda_check(cuda_req: CWLObjectType, requestCount: int) -> int:
    try:
        vmin = float(str(cuda_req["cudaVersionMin"]))
        version, devices = cuda_version_and_device_count()
        if version == "":
            # nvidia-smi not detected, or failed some other way
            return 0
        versionf = float(version)
        if versionf < vmin:
            _logger.warning("CUDA version '%s' is less than minimum version '%s'", version, vmin)
            return 0
        if requestCount > devices:
            _logger.warning("Requested %d GPU devices but only %d available", requestCount, devices)
            return 0
        return requestCount
    except Exception as e:
        _logger.warning("Error checking CUDA requirements: %s", e, exc_info=e)
        return 0



import json

from _typeshed import Incomplete
from prov.constants import *
from prov.model import Identifier as Identifier
from prov.model import Literal as Literal
from prov.model import Namespace as Namespace
from prov.model import ProvBundle as ProvBundle
from prov.model import ProvDocument as ProvDocument
from prov.model import first as first
from prov.model import parse_xsd_datetime as parse_xsd_datetime
from prov.serializers import Error as Error
from prov.serializers import Serializer as Serializer

logger: Incomplete

class ProvJSONException(Error): ...

class AnonymousIDGenerator:
    def __init__(self) -> None: ...
    def get_anon_id(self, obj, local_prefix: str = ...): ...

LITERAL_XSDTYPE_MAP: Incomplete

class ProvJSONSerializer(Serializer):
    def serialize(self, stream, **kwargs) -> None: ...
    def deserialize(self, stream, **kwargs): ...

class ProvJSONEncoder(json.JSONEncoder):
    def default(self, o): ...

class ProvJSONDecoder(json.JSONDecoder):
    def decode(self, s, *args, **kwargs): ...

def valid_qualified_name(bundle, value): ...
def encode_json_document(document): ...
def encode_json_container(bundle): ...
def decode_json_document(content, document) -> None: ...
def decode_json_container(jc, bundle) -> None: ...
def encode_json_representation(value): ...
def decode_json_representation(literal, bundle): ...
def literal_json_representation(literal): ...



"""Shared functions and other definitions."""

import collections

try:
    import fcntl
except ImportError:
    # Guard against `from .utils import ...` on windows.
    # See windows_check() in main.py
    pass
import importlib.metadata
import os
import random
import shutil
import stat
import string
import subprocess  # nosec
import sys
import tempfile
import urllib
import uuid
from collections.abc import (
    Callable,
    Generator,
    Iterable,
    MutableMapping,
    MutableSequence,
)
from datetime import datetime
from email.utils import parsedate_to_datetime
from functools import partial
from itertools import zip_longest
from pathlib import Path, PurePosixPath
from tempfile import NamedTemporaryFile
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Deque,
    Literal,
    NamedTuple,
    Optional,
    TypeAlias,
    TypedDict,
    Union,
    cast,
)

import requests
from cachecontrol import CacheControl
from cachecontrol.caches import FileCache
from mypy_extensions import mypyc_attr
from schema_salad.exceptions import ValidationException

if TYPE_CHECKING:
    from schema_salad.ref_resolver import Loader

    from .command_line_tool import CallbackJob, ExpressionJob
    from .job import CommandLineJob, JobBase
    from .stdfsaccess import StdFsAccess
    from .workflow_job import WorkflowJob

__random_outdir: str | None = None

CONTENT_LIMIT = 64 * 1024

DEFAULT_TMP_PREFIX = tempfile.gettempdir() + os.path.sep

processes_to_kill: Deque["subprocess.Popen[str]"] = collections.deque()

CWLOutputType: TypeAlias = Union[
    None,
    bool,
    str,
    int,
    float,
    MutableSequence["CWLOutputType"],
    MutableMapping[str, "CWLOutputType"],
]
CWLObjectType: TypeAlias = MutableMapping[str, Optional[CWLOutputType]]
"""Typical raw dictionary found in lightly parsed CWL."""

JobsType: TypeAlias = Union[
    "CommandLineJob", "JobBase", "WorkflowJob", "ExpressionJob", "CallbackJob"
]
JobsGeneratorType: TypeAlias = Generator[Optional[JobsType], None, None]
OutputCallbackType: TypeAlias = Callable[[Optional[CWLObjectType], str], None]
ResolverType: TypeAlias = Callable[["Loader", str], Optional[str]]
DestinationsType: TypeAlias = MutableMapping[str, Optional[CWLOutputType]]
ScatterDestinationsType: TypeAlias = MutableMapping[str, list[Optional[CWLOutputType]]]
ScatterOutputCallbackType: TypeAlias = Callable[[Optional[ScatterDestinationsType], str], None]
SinkType: TypeAlias = Union[CWLOutputType, CWLObjectType]
DirectoryType = TypedDict(
    "DirectoryType", {"class": str, "listing": list[CWLObjectType], "basename": str}
)
JSONType: TypeAlias = Union[dict[str, "JSONType"], list["JSONType"], str, int, float, bool, None]


class WorkflowStateItem(NamedTuple):
    """Workflow state item."""

    parameter: CWLObjectType
    value: CWLOutputType | None
    success: str


ParametersType: TypeAlias = list[CWLObjectType]
StepType: TypeAlias = CWLObjectType  # WorkflowStep

LoadListingType = Union[Literal["no_listing"], Literal["shallow_listing"], Literal["deep_listing"]]


def versionstring() -> str:
    """Version of CWLtool used to execute the workflow."""
    if pkg := importlib.metadata.version("cwltool"):
        return f"{sys.argv[0]} {pkg}"
    return "{} {}".format(sys.argv[0], "unknown version")


def aslist(thing: Any) -> MutableSequence[Any]:
    """Wrap any non-MutableSequence/list in a list."""
    if isinstance(thing, MutableSequence):
        return thing
    return [thing]


def copytree_with_merge(src: str, dst: str) -> None:
    if not os.path.exists(dst):
        os.makedirs(dst)
        shutil.copystat(src, dst)
    lst = os.listdir(src)
    for item in lst:
        spath = os.path.join(src, item)
        dpath = os.path.join(dst, item)
        if os.path.isdir(spath):
            copytree_with_merge(spath, dpath)
        else:
            shutil.copy2(spath, dpath)


def cmp_like_py2(dict1: dict[str, Any], dict2: dict[str, Any]) -> int:
    """
    Compare in the same manner as Python2.

    Comparison function to be used in sorting as python3 doesn't allow sorting
    of different types like str() and int().
    This function re-creates sorting nature in py2 of heterogeneous list of
    `int` and `str`
    """
    # extract lists from both dicts
    first, second = dict1["position"], dict2["position"]
    # iterate through both list till max of their size
    for i, j in zip_longest(first, second):
        if i == j:
            continue
        # in case 1st list is smaller
        # should come first in sorting
        if i is None:
            return -1
        # if 1st list is longer,
        # it should come later in sort
        elif j is None:
            return 1

        # if either of the list contains str element
        # at any index, both should be str before comparing
        if isinstance(i, str) or isinstance(j, str):
            return 1 if str(i) > str(j) else -1
        # int comparison otherwise
        return 1 if i > j else -1
    # if both lists are equal
    return 0


def bytes2str_in_dicts(
    inp: MutableMapping[str, Any] | MutableSequence[Any] | Any,
) -> str | MutableSequence[Any] | MutableMapping[str, Any]:
    """
    Convert any present byte string to unicode string, inplace.

    input is a dict of nested dicts and lists
    """
    # if input is dict, recursively call for each value
    if isinstance(inp, MutableMapping):
        for k in inp:
            inp[k] = bytes2str_in_dicts(inp[k])
        return inp

    # if list, iterate through list and fn call
    # for all its elements
    if isinstance(inp, MutableSequence):
        for idx, value in enumerate(inp):
            inp[idx] = bytes2str_in_dicts(value)
            return inp

    # if value is bytes, return decoded string,
    elif isinstance(inp, bytes):
        return str(inp, "utf-8")

    # simply return elements itself
    return inp


def visit_class(rec: Any, cls: Iterable[Any], op: Callable[..., Any]) -> None:
    """Apply a function to with "class" in cls."""
    if isinstance(rec, MutableMapping):
        if "class" in rec and rec.get("class") in cls:
            op(rec)
        for d in rec:
            visit_class(rec[d], cls, op)
    if isinstance(rec, MutableSequence):
        for d in rec:
            visit_class(d, cls, op)


def visit_field(rec: Any, field: str, op: Callable[..., Any]) -> None:
    """Apply a function to mapping with 'field'."""
    if isinstance(rec, MutableMapping):
        if field in rec:
            rec[field] = op(rec[field])
        for d in rec:
            visit_field(rec[d], field, op)
    if isinstance(rec, MutableSequence):
        for d in rec:
            visit_field(d, field, op)


def random_outdir() -> str:
    """Return the random directory name chosen to use for tool / workflow output."""
    global __random_outdir
    if not __random_outdir:
        __random_outdir = "/" + "".join(
            [random.choice(string.ascii_letters) for _ in range(6)]  # nosec
        )
        return __random_outdir
    return __random_outdir


def shared_file_lock(fd: IO[Any]) -> None:
    fcntl.flock(fd.fileno(), fcntl.LOCK_SH)


def upgrade_lock(fd: IO[Any]) -> None:
    fcntl.flock(fd.fileno(), fcntl.LOCK_EX)


def adjustFileObjs(rec: Any, op: Union[Callable[..., Any], "partial[Any]"]) -> None:
    """Apply an update function to each File object in the object `rec`."""
    visit_class(rec, ("File",), op)


def adjustDirObjs(rec: Any, op: Union[Callable[..., Any], "partial[Any]"]) -> None:
    """Apply an update function to each Directory object in the object `rec`."""
    visit_class(rec, ("Directory",), op)


def dedup(listing: list[CWLObjectType]) -> list[CWLObjectType]:
    marksub = set()

    def mark(d: dict[str, str]) -> None:
        marksub.add(d["location"])

    for entry in listing:
        if entry["class"] == "Directory":
            for e in cast(list[CWLObjectType], entry.get("listing", [])):
                adjustFileObjs(e, mark)
                adjustDirObjs(e, mark)

    dd = []
    markdup: set[str] = set()
    for r in listing:
        if r["location"] not in marksub and r["location"] not in markdup:
            dd.append(r)
            markdup.add(cast(str, r["location"]))

    return dd


def get_listing(fs_access: "StdFsAccess", rec: CWLObjectType, recursive: bool = True) -> None:
    """Expand, recursively, any 'listing' fields in a Directory."""
    if rec.get("class") != "Directory":
        finddirs: list[CWLObjectType] = []
        visit_class(rec, ("Directory",), finddirs.append)
        for f in finddirs:
            get_listing(fs_access, f, recursive=recursive)
        return
    if "listing" in rec:
        return
    listing: list[CWLOutputType] = []
    loc = cast(str, rec["location"])
    for ld in fs_access.listdir(loc):
        parse = urllib.parse.urlparse(ld)
        bn = os.path.basename(urllib.request.url2pathname(parse.path))
        if fs_access.isdir(ld):
            ent: MutableMapping[str, Any] = {
                "class": "Directory",
                "location": ld,
                "basename": bn,
            }
            if recursive:
                get_listing(fs_access, ent, recursive)
            listing.append(ent)
        else:
            listing.append({"class": "File", "location": ld, "basename": bn})
    rec["listing"] = listing


def trim_listing(obj: dict[str, Any]) -> None:
    """
    Remove 'listing' field from Directory objects that are file references.

    It redundant and potentially expensive to pass fully enumerated Directory
    objects around if not explicitly needed, so delete the 'listing' field when
    it is safe to do so.
    """
    if obj.get("location", "").startswith("file://") and "listing" in obj:
        del obj["listing"]


def downloadHttpFile(httpurl: str) -> tuple[str, datetime | None]:
    """
    Download a remote file, possibly using a locally cached copy.

    Returns a tuple:
    - the local path for the downloaded file
    - the Last-Modified timestamp if received from the remote server.
    """
    cache_session = None
    if "XDG_CACHE_HOME" in os.environ:
        directory = os.environ["XDG_CACHE_HOME"]
    elif "HOME" in os.environ:
        directory = os.path.join(os.environ["HOME"], ".cache")
    else:
        directory = os.path.join(os.path.expanduser("~"), ".cache")

    cache_session = CacheControl(
        requests.Session(),
        cache=FileCache(os.path.join(directory, "cwltool")),
    )

    r = cache_session.get(httpurl, stream=True)
    with NamedTemporaryFile(mode="wb", delete=False) as f:
        for chunk in r.iter_content(chunk_size=16384):
            if chunk:  # filter out keep-alive new chunks
                f.write(chunk)
    r.close()

    date_raw: str | None = r.headers.get("Last-Modified", None)
    date: datetime | None = parsedate_to_datetime(date_raw) if date_raw else None
    if date:
        date_epoch = date.timestamp()
        os.utime(f.name, (date_epoch, date_epoch))

    return str(f.name), date


def ensure_writable(path: str, include_root: bool = False) -> None:
    """
    Ensure that 'path' is writable.

    If 'path' is a directory, then all files and directories under 'path' are
    made writable, recursively. If 'path' is a file or if 'include_root' is
    `True`, then 'path' itself is made writable.
    """

    def add_writable_flag(p: str) -> None:
        st = os.stat(p)
        mode = stat.S_IMODE(st.st_mode)
        os.chmod(p, mode | stat.S_IWUSR)

    if os.path.isdir(path):
        if include_root:
            add_writable_flag(path)
        for root, dirs, files_ in os.walk(path):
            for name in files_:
                add_writable_flag(os.path.join(root, name))
            for name in dirs:
                add_writable_flag(os.path.join(root, name))
    else:
        add_writable_flag(path)


def ensure_non_writable(path: str) -> None:
    """Attempt to change permissions to ensure that a path is not writable."""
    if os.path.isdir(path):
        for root, dirs, files_ in os.walk(path):
            for name in files_:
                j = os.path.join(root, name)
                st = os.stat(j)
                mode = stat.S_IMODE(st.st_mode)
                os.chmod(j, mode & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH)
            for name in dirs:
                j = os.path.join(root, name)
                st = os.stat(j)
                mode = stat.S_IMODE(st.st_mode)
                os.chmod(j, mode & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH)
    else:
        st = os.stat(path)
        mode = stat.S_IMODE(st.st_mode)
        os.chmod(path, mode & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH)


def normalizeFilesDirs(
    job: None | (
        MutableSequence[MutableMapping[str, Any]] | MutableMapping[str, Any] | DirectoryType
    ),
) -> None:
    def addLocation(d: dict[str, Any]) -> None:
        if "location" not in d:
            if d["class"] == "File" and ("contents" not in d):
                raise ValidationException(
                    "Anonymous file object must have 'contents' and 'basename' fields."
                )
            if d["class"] == "Directory" and ("listing" not in d or "basename" not in d):
                raise ValidationException(
                    "Anonymous directory object must have 'listing' and 'basename' fields."
                )
            d["location"] = "_:" + str(uuid.uuid4())
            if "basename" not in d:
                d["basename"] = d["location"][2:]

        parse = urllib.parse.urlparse(d["location"])
        path = parse.path
        # strip trailing slash
        if path.endswith("/"):
            if d["class"] != "Directory":
                raise ValidationException(
                    "location '%s' ends with '/' but is not a Directory" % d["location"]
                )
            path = path.rstrip("/")
            d["location"] = urllib.parse.urlunparse(
                (
                    parse.scheme,
                    parse.netloc,
                    path,
                    parse.params,
                    parse.query,
                    parse.fragment,
                )
            )

        if not d.get("basename"):
            if path.startswith("_:"):
                d["basename"] = str(path[2:])
            else:
                d["basename"] = str(os.path.basename(urllib.request.url2pathname(path)))

        if d["class"] == "File":
            nr, ne = os.path.splitext(d["basename"])
            if d.get("nameroot") != nr:
                d["nameroot"] = str(nr)
            if d.get("nameext") != ne:
                d["nameext"] = str(ne)

    visit_class(job, ("File", "Directory"), addLocation)


def posix_path(local_path: str) -> str:
    return str(PurePosixPath(Path(local_path)))


def local_path(posix_path: str) -> str:
    return str(Path(posix_path))


def create_tmp_dir(tmpdir_prefix: str) -> str:
    """Create a temporary directory that respects the given tmpdir_prefix."""
    tmp_dir, tmp_prefix = os.path.split(tmpdir_prefix)
    return tempfile.mkdtemp(prefix=tmp_prefix, dir=tmp_dir)


@mypyc_attr(allow_interpreted_subclasses=True)
class HasReqsHints:
    """Base class for get_requirement()."""

    def __init__(self) -> None:
        """Initialize this reqs decorator."""
        self.requirements: list[CWLObjectType] = []
        self.hints: list[CWLObjectType] = []

    def get_requirement(self, feature: str) -> tuple[CWLObjectType | None, bool | None]:
        """Retrieve the named feature from the requirements field, or the hints field."""
        for item in reversed(self.requirements):
            if item["class"] == feature:
                return (item, True)
        for item in reversed(self.hints):
            if item["class"] == feature:
                return (item, False)
        return (None, None)



"""Shared context objects that replace use of kwargs."""

import copy
import os
import shutil
import tempfile
import threading
from collections.abc import Callable, Iterable
from typing import IO, TYPE_CHECKING, Any, Literal, Optional, TextIO, Union

from ruamel.yaml.comments import CommentedMap
from schema_salad.avro.schema import Names
from schema_salad.ref_resolver import Loader
from schema_salad.utils import FetcherCallableType

from .mpi import MpiConfig
from .pathmapper import PathMapper
from .stdfsaccess import StdFsAccess
from .utils import DEFAULT_TMP_PREFIX, CWLObjectType, HasReqsHints, ResolverType

if TYPE_CHECKING:
    from _typeshed import SupportsWrite
    from cwl_utils.parser.cwl_v1_2 import LoadingOptions

    from .builder import Builder
    from .cwlprov.provenance_profile import ProvenanceProfile
    from .cwlprov.ro import ResearchObject
    from .mutation import MutationManager
    from .process import Process
    from .secrets import SecretStore
    from .software_requirements import DependenciesConfiguration
    from .workflow_job import WorkflowJobStep


class ContextBase:
    """Shared kwargs based initializer for :py:class:`RuntimeContext` and :py:class:`LoadingContext`."""

    def __init__(self, kwargs: dict[str, Any] | None = None) -> None:
        """Initialize."""
        if kwargs:
            for k, v in kwargs.items():
                if hasattr(self, k):
                    setattr(self, k, v)


def make_tool_notimpl(toolpath_object: CommentedMap, loadingContext: "LoadingContext") -> "Process":
    """Fake implementation of the make tool function."""
    raise NotImplementedError()


default_make_tool = make_tool_notimpl


def log_handler(
    outdir: str,
    base_path_logs: str,
    stdout_path: str | None,
    stderr_path: str | None,
) -> None:
    """Move logs from log location to final output."""
    if outdir != base_path_logs:
        if stdout_path:
            new_stdout_path = stdout_path.replace(base_path_logs, outdir)
            shutil.copy2(stdout_path, new_stdout_path)
        if stderr_path:
            new_stderr_path = stderr_path.replace(base_path_logs, outdir)
            shutil.copy2(stderr_path, new_stderr_path)


def set_log_dir(outdir: str, log_dir: str, subdir_name: str) -> str:
    """Set the log directory."""
    if log_dir == "":
        return outdir
    else:
        return log_dir + "/" + subdir_name


class LoadingContext(ContextBase):
    def __init__(self, kwargs: dict[str, Any] | None = None) -> None:
        """Initialize the LoadingContext from the kwargs."""
        self.debug: bool = False
        self.metadata: CWLObjectType = {}
        self.requirements: list[CWLObjectType] | None = None
        self.hints: list[CWLObjectType] | None = None
        self.overrides_list: list[CWLObjectType] = []
        self.loader: Loader | None = None
        self.avsc_names: Names | None = None
        self.disable_js_validation: bool = False
        self.js_hint_options_file: str | None = None
        self.do_validate: bool = True
        self.enable_dev: bool = False
        self.strict: bool = True
        self.resolver: ResolverType | None = None
        self.fetcher_constructor: FetcherCallableType | None = None
        self.construct_tool_object = default_make_tool
        self.research_obj: ResearchObject | None = None
        self.orcid: str = ""
        self.cwl_full_name: str = ""
        self.host_provenance: bool = False
        self.user_provenance: bool = False
        self.prov_obj: Optional["ProvenanceProfile"] = None
        self.do_update: bool | None = None
        self.jobdefaults: CommentedMap | None = None
        self.doc_cache: bool = True
        self.relax_path_checks: bool = False
        self.singularity: bool = False
        self.podman: bool = False
        self.eval_timeout: float = 60
        self.codegen_idx: dict[str, tuple[Any, "LoadingOptions"]] = {}
        self.fast_parser = False
        self.skip_resolve_all = False
        self.skip_schemas = False

        super().__init__(kwargs)

    def copy(self) -> "LoadingContext":
        """Return a copy of this :py:class:`LoadingContext`."""
        return copy.copy(self)


class RuntimeContext(ContextBase):
    outdir: str | None = None
    tmpdir: str = ""
    tmpdir_prefix: str = DEFAULT_TMP_PREFIX
    tmp_outdir_prefix: str = ""
    stagedir: str = ""

    def __init__(self, kwargs: dict[str, Any] | None = None) -> None:
        """Initialize the RuntimeContext from the kwargs."""
        select_resources_callable = Callable[
            [dict[str, Union[int, float]], RuntimeContext],
            dict[str, Union[int, float]],
        ]
        self.user_space_docker_cmd: str | None = None
        self.secret_store: Optional["SecretStore"] = None
        self.no_read_only: bool = False
        self.custom_net: str | None = None
        self.no_match_user: bool = False
        self.preserve_environment: Iterable[str] | None = None
        self.preserve_entire_environment: bool = False
        self.use_container: bool = True
        self.force_docker_pull: bool = False

        self.rm_tmpdir: bool = True
        self.pull_image: bool = True
        self.rm_container: bool = True
        self.move_outputs: Literal["move"] | Literal["leave"] | Literal["copy"] = "move"
        self.log_dir: str = ""
        self.set_log_dir = set_log_dir
        self.log_dir_handler = log_handler
        self.streaming_allowed: bool = False

        self.singularity: bool = False
        self.image_base_path: str | None = None
        self.podman: bool = False
        self.debug: bool = False
        self.compute_checksum: bool = True
        self.name: str = ""
        self.default_container: str | None = ""
        self.find_default_container: Callable[[HasReqsHints], str | None] | None = None
        self.cachedir: str | None = None
        self.part_of: str = ""
        self.basedir: str = ""
        self.toplevel: bool = False
        self.mutation_manager: Optional["MutationManager"] = None
        self.make_fs_access = StdFsAccess
        self.path_mapper = PathMapper
        self.builder: Optional["Builder"] = None
        self.docker_outdir: str = ""
        self.docker_tmpdir: str = ""
        self.docker_stagedir: str = ""
        self.js_console: bool = False
        self.job_script_provider: DependenciesConfiguration | None = None
        self.select_resources: select_resources_callable | None = None
        self.eval_timeout: float = 60
        self.postScatterEval: Callable[[CWLObjectType], CWLObjectType | None] | None = None
        self.on_error: Literal["stop"] | Literal["continue"] = "stop"
        self.strict_memory_limit: bool = False
        self.strict_cpu_limit: bool = False
        self.cidfile_dir: str | None = None
        self.cidfile_prefix: str | None = None

        self.workflow_eval_lock: Union[threading.Condition, None] = None
        self.research_obj: ResearchObject | None = None
        self.orcid: str = ""
        self.cwl_full_name: str = ""
        self.process_run_id: str | None = None
        self.prov_host: bool = False
        self.prov_user: bool = False
        self.prov_obj: ProvenanceProfile | None = None
        self.mpi_config: MpiConfig = MpiConfig()
        self.default_stdout: IO[bytes] | TextIO | None = None
        self.default_stderr: IO[bytes] | TextIO | None = None
        self.validate_only: bool = False
        self.validate_stdout: Optional["SupportsWrite[str]"] = None
        self.workflow_job_step_name_callback: None | (
            Callable[[WorkflowJobStep, CWLObjectType], str]
        ) = None

        super().__init__(kwargs)
        if self.tmp_outdir_prefix == "":
            self.tmp_outdir_prefix = self.tmpdir_prefix

    def get_outdir(self) -> str:
        """Return :py:attr:`outdir` or create one with :py:attr:`tmp_outdir_prefix`."""
        if self.outdir:
            return self.outdir
        return self.create_outdir()

    def get_tmpdir(self) -> str:
        """Return :py:attr:`tmpdir` or create one with :py:attr:`tmpdir_prefix`."""
        if self.tmpdir:
            return self.tmpdir
        return self.create_tmpdir()

    def get_stagedir(self) -> str:
        """Return :py:attr:`stagedir` or create one with :py:attr:`tmpdir_prefix`."""
        if self.stagedir:
            return self.stagedir
        tmp_dir, tmp_prefix = os.path.split(self.tmpdir_prefix)
        return tempfile.mkdtemp(prefix=tmp_prefix, dir=tmp_dir)

    def create_tmpdir(self) -> str:
        """Create a temporary directory that respects :py:attr:`tmpdir_prefix`."""
        tmp_dir, tmp_prefix = os.path.split(self.tmpdir_prefix)
        return tempfile.mkdtemp(prefix=tmp_prefix, dir=tmp_dir)

    def create_outdir(self) -> str:
        """Create a temporary directory that respects :py:attr:`tmp_outdir_prefix`."""
        out_dir, out_prefix = os.path.split(self.tmp_outdir_prefix)
        return tempfile.mkdtemp(prefix=out_prefix, dir=out_dir)

    def copy(self) -> "RuntimeContext":
        """Return a copy of this :py:class:`RuntimeContext`."""
        return copy.copy(self)


def getdefault(val: Any, default: Any) -> Any:
    """Return the ``val`` using the ``default`` as backup in case the val is ``None``."""
    if val is None:
        return default
    else:
        return val



"""This module handles resolution of SoftwareRequirement hints.

This is accomplished mainly by adapting cwltool internals to galaxy-tool-util's
concept of "dependencies". Despite the name, galaxy-tool-util is a light weight
library that can be used to map SoftwareRequirements in all sorts of ways -
Homebrew, Conda, custom scripts, environment modules. We'd be happy to find
ways to adapt new packages managers and such as well.
"""

import argparse
import importlib.metadata
import os
import string
from collections.abc import MutableMapping, MutableSequence
from typing import TYPE_CHECKING, Any, Union, cast

from .utils import HasReqsHints

if TYPE_CHECKING:
    from galaxy.tool_util.deps.requirements import ToolRequirements

    from .builder import Builder

try:
    importlib.metadata.Distribution.from_name("galaxy-tool-util")
    SOFTWARE_REQUIREMENTS_ENABLED = True
except ModuleNotFoundError:
    SOFTWARE_REQUIREMENTS_ENABLED = False

COMMAND_WITH_DEPENDENCIES_TEMPLATE = string.Template("""#!/bin/bash
cat > modify_environment.bash <<'EOF'
$handle_dependencies
# First try env -0
if ! env -0 > "output_environment.dat" 2> /dev/null; then
    # If that fails, use the python script.
    # In some circumstances (see PEP 538) this will the add LC_CTYPE env var.
    python3 "env_to_stdout.py" > "output_environment.dat"
fi
EOF
python3 "run_job.py" "job.json" "modify_environment.bash"
""")


class DependenciesConfiguration:
    """Dependency configuration class, for RuntimeContext.job_script_provider."""

    def __init__(self, args: argparse.Namespace) -> None:
        """Initialize."""
        self.tool_dependency_dir: str | None = None
        self.dependency_resolvers_config_file: str | None = None
        conf_file = getattr(args, "beta_dependency_resolvers_configuration", None)
        tool_dependency_dir = getattr(args, "beta_dependencies_directory", None)
        conda_dependencies = getattr(args, "beta_conda_dependencies", None)
        if conf_file is not None and os.path.exists(conf_file):
            self.use_tool_dependencies = True
            if tool_dependency_dir is None:
                tool_dependency_dir = os.path.abspath(os.path.dirname(conf_file))
            self.tool_dependency_dir = tool_dependency_dir
            self.dependency_resolvers_config_file = os.path.abspath(conf_file)
        elif conda_dependencies is not None:
            if tool_dependency_dir is None:
                tool_dependency_dir = os.path.abspath("./cwltool_deps")
            self.tool_dependency_dir = tool_dependency_dir
            self.use_tool_dependencies = True
            self.dependency_resolvers_config_file = None
        else:
            self.use_tool_dependencies = False
        if self.tool_dependency_dir and not os.path.exists(self.tool_dependency_dir):
            os.makedirs(self.tool_dependency_dir)

    def build_job_script(self, builder: "Builder", command: list[str]) -> str:
        """Use the galaxy-tool-util library to construct a build script."""
        from galaxy.tool_util import deps

        ensure_galaxy_lib_available()
        resolution_config_dict = {
            "use": self.use_tool_dependencies,
            "default_base_path": self.tool_dependency_dir,
        }
        app_config = {
            "conda_auto_install": True,
            "conda_auto_init": True,
            "debug": builder.debug,
        }
        tool_dependency_manager: "deps.DependencyManager" = deps.build_dependency_manager(
            app_config_dict=app_config,
            resolution_config_dict=resolution_config_dict,
            conf_file=self.dependency_resolvers_config_file,
        )
        handle_dependencies: str = ""
        if dependencies := get_dependencies(builder):
            handle_dependencies = "\n".join(
                tool_dependency_manager.dependency_shell_commands(
                    dependencies, job_directory=builder.tmpdir
                )
            )

        template_kwds: dict[str, str] = dict(handle_dependencies=handle_dependencies)
        job_script = COMMAND_WITH_DEPENDENCIES_TEMPLATE.substitute(template_kwds)
        return job_script


def get_dependencies(
    builder: HasReqsHints,
) -> "ToolRequirements":
    software_requirement, _ = builder.get_requirement("SoftwareRequirement")
    from galaxy.tool_util.deps.requirements import ToolRequirement, ToolRequirements

    dependencies: list[Union["ToolRequirement", dict[str, Any]]] = []
    if software_requirement and software_requirement.get("packages"):
        packages = cast(
            MutableSequence[MutableMapping[str, Union[str, MutableSequence[str]]]],
            software_requirement.get("packages"),
        )
        for package in packages:
            version = package.get("version", None)
            if isinstance(version, MutableSequence):
                if version:
                    version = version[0]
                else:
                    version = None
            specs = [{"uri": s} for s in package.get("specs", [])]
            dependencies.append(
                ToolRequirement.from_dict(
                    dict(
                        name=cast(str, package["package"]).split("#")[-1],
                        version=version,
                        type="package",
                        specs=specs,
                    )
                )
            )

    return ToolRequirements.from_list(dependencies)


def get_container_from_software_requirements(
    use_biocontainers: bool, builder: HasReqsHints, container_image_cache_path: str | None = "."
) -> str | None:
    if use_biocontainers:
        ensure_galaxy_lib_available()
        from galaxy.tool_util.deps.container_classes import DOCKER_CONTAINER_TYPE
        from galaxy.tool_util.deps.containers import ContainerRegistry
        from galaxy.tool_util.deps.dependencies import AppInfo, ToolInfo

        app_info: AppInfo = AppInfo(
            involucro_auto_init=True,
            enable_mulled_containers=True,
            container_image_cache_path=container_image_cache_path,
        )
        container_registry: ContainerRegistry = ContainerRegistry(app_info)
        requirements = get_dependencies(builder)
        tool_info: ToolInfo = ToolInfo(requirements=requirements)
        container_description = container_registry.find_best_container_description(
            [DOCKER_CONTAINER_TYPE], tool_info
        )
        if container_description:
            return str(container_description.identifier)

    return None


def ensure_galaxy_lib_available() -> None:
    if not SOFTWARE_REQUIREMENTS_ENABLED:
        raise Exception(
            "Optional Python library galaxy-tool-util not available, it is required for this configuration."
        )
